from typing import Generator
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type
from typing import Union
from typing import cast

from pytezos.context.abstract import AbstractContext
from pytezos.michelson.micheline import Micheline
from pytezos.michelson.micheline import parse_micheline_value
from pytezos.michelson.types.adt import ADTMixin
from pytezos.michelson.types.adt import Nested
from pytezos.michelson.types.adt import wrap_or
from pytezos.michelson.types.base import MichelsonType
from pytezos.michelson.types.base import Undefined
from pytezos.michelson.types.base import undefined
from pytezos.michelson.types.core import Unit
from pytezos.michelson.types.core import UnitType


class LeftLiteral(Micheline, prim='Left', args_len=1):
    pass


class RightLiteral(Micheline, prim='Right', args_len=1):
    pass


class OrType(MichelsonType, ADTMixin, prim='or', args_len=2):
    is_enum: bool

    def __init__(self, items: Tuple[Union[undefined, MichelsonType], ...]):
        super(OrType, self).__init__()
        self.items = items

    def __repr__(self):
        return f'({" + ".join(map(repr, self.items))})'

    def __iter__(self) -> Generator[Optional[MichelsonType], None, None]:
        yield from self.items  # type: ignore

    def __eq__(self, other):  # type: ignore
        if not isinstance(other, OrType):
            return False
        return all(item == other.items[i] for i, item in enumerate(self.items))

    def __lt__(self, other: 'OrType'):  # type: ignore
        if self.is_left() and other.is_right():
            return True
        elif self.is_left() and other.is_left():
            return self.items[0] < other.items[0]  # type: ignore
        elif self.is_right() and other.is_right():
            return self.items[1] < other.items[1]  # type: ignore
        else:
            return False

    def __hash__(self):
        return hash(self.items)

    def is_left(self) -> bool:
        return isinstance(self.items[0], MichelsonType)

    def is_right(self) -> bool:
        return isinstance(self.items[1], MichelsonType)

    def resolve(self) -> MichelsonType:
        return next(item for item in self if isinstance(item, MichelsonType))

    @staticmethod
    def from_left(left: MichelsonType, right_type: Type[MichelsonType]):
        cls = OrType.create_type(args=[type(left), right_type])
        return cls((left, Undefined))

    @staticmethod
    def from_right(right: MichelsonType, left_type: Type[MichelsonType]):
        cls = OrType.create_type(args=[left_type, type(right)])
        return cls((Undefined, right))

    @classmethod
    def iter_type_args(cls, entrypoints=False, path='') -> Generator[Tuple[str, Type[MichelsonType]], None, None]:
        for i, arg in enumerate(cls.args):
            if issubclass(arg, OrType):
                if entrypoints and arg.field_name:
                    yield path + str(i), arg
                yield from arg.iter_type_args(entrypoints=entrypoints, path=path + str(i))  # type: ignore
            elif entrypoints is False or arg.field_name:
                yield path + str(i), arg

    @classmethod
    def create_type(
        cls,
        args: List[Type['Micheline']],
        annots: Optional[list] = None,
        **kwargs,
    ) -> Type['OrType']:
        def all_units(arguments: List[Type['Micheline']]):
            for arg in arguments:
                if issubclass(arg, OrType):
                    if not all_units(arg.args):
                        return False
                elif not issubclass(arg, UnitType):
                    return False
            return True

        is_enum = all_units(args)
        res = super(OrType, cls).create_type(args=args, annots=annots, is_enum=is_enum, **kwargs)
        return cast(Type['OrType'], res)

    @classmethod
    def generate_pydoc(cls, definitions: list, inferred_name=None, comparable=False):
        name = cls.field_name or cls.type_name or inferred_name or f'{cls.prim}_{len(definitions)}'
        flat_args = cls.get_flat_args(infer_names=True)
        assert isinstance(flat_args, dict), f'sum type has to be named (in the scope of PyTezos)'
        if cls.is_enum:
            doc = ' || '.join(flat_args.keys())
        else:
            variants = [
                (entrypoint, arg.generate_pydoc(definitions, inferred_name=entrypoint))
                for entrypoint, arg in flat_args.items()
            ]
            if comparable:
                doc = ' ||\n\t'.join(f'( "{entrypoint}", {arg_doc} )' for entrypoint, arg_doc in variants)
            else:
                doc = ' ||\n\t'.join(f'{{ "{entrypoint}": {arg_doc} }}' for entrypoint, arg_doc in variants)
        definitions.insert(0, (name, doc))
        return f'${name}'

    @classmethod
    def dummy(cls, context: AbstractContext):
        return cls((cls.args[0].dummy(context), Undefined))

    @classmethod
    def from_micheline_value(cls, val_expr) -> 'OrType':
        value = parse_micheline_value(
            val_expr,
            {
                ('Left', 1): lambda x: (cls.args[0].from_micheline_value(x[0]), Undefined),
                ('Right', 1): lambda x: (Undefined, cls.args[1].from_micheline_value(x[0])),
            },
        )
        return cls(value)

    @classmethod
    def from_python_object(cls, py_obj) -> 'OrType':
        if isinstance(py_obj, list):
            py_obj = tuple(py_obj)
        elif isinstance(py_obj, str):
            assert cls.is_enum, 'string values allowed for enums only'
            py_obj = {py_obj: Unit}
        elif isinstance(py_obj, tuple):
            assert len(py_obj) == 2, f'expected `(entrypoint, value)`, got {py_obj}'
            py_obj = {py_obj[0]: py_obj[1]}

        if isinstance(py_obj, dict):
            assert len(py_obj) == 1, f'single key expected, got {len(py_obj)}'
            entrypoint = next(iter(py_obj))
            _, key_to_path, _ = cls.get_type_layout(infer_names=True)
            assert key_to_path, f'sum type has to be named (in the scope of PyTezos)'
            return cls.from_python_object(wrap_or(py_obj[entrypoint], key_to_path[entrypoint]))
        elif isinstance(py_obj, Nested):
            value = tuple(
                Undefined if py_obj[i] is Undefined else cls.args[i].from_python_object(py_obj[i]) for i in [0, 1]
            )
            return cls(value)
        else:
            raise AssertionError(f'expected list, tuple, or dict, got `{py_obj}`')

    def iter_values(self, path='') -> Generator[Tuple[str, MichelsonType], None, None]:
        for i, arg in enumerate(self.items):
            if isinstance(arg, OrType):
                yield from arg.iter_values(path + str(i))
            elif isinstance(arg, MichelsonType):
                yield path + str(i), arg
            else:
                assert arg == Undefined, f'expected Michelson type or undefined, got {arg}'

    def to_literal(self) -> Type[Micheline]:
        if self.is_left():
            return LeftLiteral.create_type(args=[self.items[0].to_literal()])  # type: ignore
        else:
            return RightLiteral.create_type(args=[self.items[1].to_literal()])  # type: ignore

    def to_micheline_value(self, mode='readable', lazy_diff=False):
        for i, prim in enumerate(['Left', 'Right']):
            if isinstance(self.items[i], MichelsonType):
                return {'prim': prim, 'args': [self.items[i].to_micheline_value(mode=mode, lazy_diff=lazy_diff)]}
        raise AssertionError(f'unexpected value {self.items}')

    def to_python_object(self, try_unpack=False, lazy_diff=False, comparable=False) -> Union[tuple, dict]:
        flat_values = self.get_flat_values(infer_names=True)
        assert (
            isinstance(flat_values, dict) and len(flat_values) == 1
        ), f'sum type has to be named (in the scope of PyTezos)'
        entrypoint = next(iter(flat_values))
        if self.is_enum:
            return entrypoint  # type: ignore
        else:
            py_obj = flat_values[entrypoint].to_python_object(
                try_unpack=try_unpack,
                lazy_diff=lazy_diff,
                comparable=comparable,
            )
            return (entrypoint, py_obj) if comparable else {entrypoint: py_obj}

    def merge_lazy_diff(self, lazy_diff: List[dict]) -> 'OrType':
        items = tuple(
            item.merge_lazy_diff(lazy_diff) if isinstance(item, MichelsonType) else item for item in self.items
        )
        return type(self)(items)  # type: ignore

    def aggregate_lazy_diff(self, lazy_diff: List[dict], mode='readable'):
        items = tuple(
            item.aggregate_lazy_diff(lazy_diff, mode=mode) if isinstance(item, MichelsonType) else item
            for item in self.items
        )
        return type(self)(items)  # type: ignore

    def attach_context(self, context: AbstractContext, big_map_copy=False):
        for item in self:
            if isinstance(item, MichelsonType):
                item.attach_context(context, big_map_copy=big_map_copy)

    def __getitem__(self, key: Union[int, str]) -> MichelsonType:
        return self.get_value(key, infer_names=True)



from typing import Any
from typing import Callable
from typing import Generator
from typing import Tuple

from pytezos.crypto.encoding import is_bh
from pytezos.jupyter import get_attr_docstring
from pytezos.logging import logger
from pytezos.rpc.node import RpcError
from pytezos.rpc.query import RpcQuery


def find_state_change_intervals(
    head: int,
    last: int,
    get: Callable,
    equals: Callable,
    step=60,
) -> Generator:
    succ_value = get(head)
    logger.debug('%s at head %s' % succ_value, head)

    for level in range(head - step, last, -step):
        value = get(level)
        logger.debug('%s at level %s' % value, level)

        if not equals(value, succ_value):
            logger.debug('%s -> %s at (%s, {level + step})' % value, succ_value, level)
            yield level + step, succ_value, level, value
            succ_value = value


def find_state_change(
    head: int,
    last: int,
    get: Callable,
    equals: Callable,
    pred_value: Any,
) -> (int, Any):  # type: ignore
    def bisect(start: int, end: int):
        if end == start + 1:
            return end, get(end)

        level = (end + start) // 2
        value = get(level)
        logger.debug('%s at level %s' % value, level)

        if equals(value, pred_value):
            return bisect(level, end)
        else:
            return bisect(start, level)

    return bisect(last, head)


def walk_state_change_interval(
    head: int,
    last: int,
    get: Callable,
    equals: Callable,
    head_value: Any,
    last_value: Any,
) -> Generator:
    level = last
    value = last_value
    while not equals(value, head_value):
        level, value = find_state_change(head, level, get, equals, pred_value=value)
        logger.debug('%s -> %s at %s' % last_value, value, level)
        yield level, value


def find_state_changes(
    head: int,
    last: int,
    get: Callable,
    equals: Callable,
    step=60,
) -> Generator:
    state_change_intervals = find_state_change_intervals(head, last, get, equals, step)
    for int_head, int_head_value, int_tail, int_last_value in state_change_intervals:
        yield from walk_state_change_interval(
            int_head,
            int_tail,
            get,
            equals,
            head_value=int_head_value,
            last_value=int_last_value,
        )


class BlockSliceQuery(RpcQuery):
    def __init__(self, start: int, stop=None, **kwargs):
        super().__init__(**kwargs)
        self._start: int = start
        self._stop = stop or 'head'

    def __repr__(self):
        res = [
            super().__repr__(),
            f'\nBlock range\n`{self._start}` — `{self._stop}`',
            f'\n(){get_attr_docstring(BlockSliceQuery, "__call__")}',
        ]
        return '\n'.join(res)

    def __getitem__(self, item):
        """Get block by index.

        :param item: Index inside given block range
        :rtype: BlockQuery
        """
        start, stop = self.get_range()
        if item >= 0:
            return self._getitem(start + item)
        else:
            return self._getitem(stop + item + 1)

    def __call__(self) -> list:  # type: ignore
        """Get block hashes (base58) for this interval."""
        header = self._getitem(self._stop).header()
        if is_bh(self._stop):
            head = self._stop
        else:
            head = header['hash']

        if self._start < 0:
            length = abs(self._start)
        else:
            length = header['level'] - self._start + 1

        return super().__call__(length=min(header['level'], length), head=head)

    def get_range(self):
        """Get block level range."""

        def get_level(x):
            if isinstance(x, int):
                if x < 0:
                    return max(0, self.head.level() + x)
                elif x > 0:
                    return x
                else:
                    return 1
            else:
                return self._getitem(x).header()['level']

        return get_level(self._start), get_level(self._stop)

    def find_proposal_injection(self, proposal_id):
        """Find proposal injection.

        :param proposal_id: Proposal hash (base58)
        """
        last, head = self.get_range()
        level, _ = find_state_change(
            head=head - 1,  # proposals are empty at the last block
            last=last,
            get=lambda x: self._getitem(x).votes.proposals[proposal_id](),
            equals=lambda x, y: x == y,
            pred_value=0,
        )
        votes = self._getitem(level).operations.find_votes(proposal_id)
        assert len(votes) == 1
        return votes

    def find_upvotes(self, proposal_id) -> Generator:
        """Find upvoting operations for the given proposal.

        :param proposal_id: Proposal hash (base58)
        :returns: Generator (lazy)
        """
        last, head = self.get_range()
        state_changes = find_state_changes(
            head=head - 1,  # proposals are empty at the last block
            last=last,
            get=lambda x: self._getitem(x).votes.proposals[proposal_id](),
            equals=lambda x, y: x == y,
        )
        for level, _ in state_changes:
            yield from self._getitem(level).operations.find_upvotes(proposal_id)

    def find_ballots(self) -> Generator:
        """Find ballot operations for the current period.

        :returns: Generator (lazy)
        """
        last, head = self.get_range()
        state_changes = find_state_changes(
            head=head - 1,  # ballots are empty at the last block
            last=last,
            get=lambda x: self._getitem(x).votes.ballots(),
            equals=lambda x, y: x == y,
        )
        for level, _ in state_changes:
            yield from self._getitem(level).operations.find_ballots()

    def find_origination(self, contract_id):
        """Find contract origination.

        :param contract_id: Contract ID (KT-address)
        """

        def get_counter(x):
            try:
                return self._getitem(x).context.contracts[contract_id].counter()
            except RpcError:
                return None

        level, _ = find_state_change(
            head=self.head.level(),
            last=0,
            get=get_counter,
            equals=lambda x, y: x == y,
            pred_value=None,
        )
        return self._getitem(level).operations.find_origination(contract_id)

    def find_operation(self, operation_group_hash) -> dict:
        """Find operation by hash.

        :param operation_group_hash: base58
        :raises: StopIteration if not found
        """
        last, head = self.get_range()
        if self._start < 0:
            levels = range(head, max(1, last - 1), -1)
        else:
            levels = range(last, head + 1, 1)

        for block_level in levels:
            logger.debug(f'Looking for operation %s in block %s...' % (operation_group_hash, block_level))
            try:
                return self._getitem(block_level).operations[operation_group_hash]()
            except StopIteration:
                continue

        raise StopIteration(operation_group_hash)


class PeriodQuery(RpcQuery):
    __pos_key__ = ''
    __val_key__ = ''

    def _get_item(self, item) -> BlockSliceQuery:
        lvl = self.head.metadata()['level_info']
        blocks_per_period = int((lvl['level'] - lvl[self.__pos_key__] - 1) / lvl[self.__val_key__])

        def get_range(x):
            if x >= 0:
                start_lvl = (max(1, x) - 1) * blocks_per_period + 1
            else:
                start_lvl = (lvl[self.__val_key__] + x + 1) * blocks_per_period + 1

            stop_lvl = start_lvl + blocks_per_period - 1
            if stop_lvl > lvl['level']:
                stop_lvl = None

            return start_lvl, stop_lvl

        if isinstance(item, slice):
            start, _ = get_range(item.start or 1)
            _, stop = get_range(item.stop or -1)
        elif isinstance(item, int):
            start, stop = get_range(item)
        else:
            raise NotImplementedError(item)

        return BlockSliceQuery(
            start=start,
            stop=stop,
            node=self.node,
            path=self._wild_path,
            params=self._params,
        )


class CyclesQuery(PeriodQuery):
    __pos_key__ = 'cycle_position'
    __val_key__ = 'cycle'

    def __call__(self, **params):
        """Get current cycle."""
        return self.head.cycle()

    def __getitem__(self, item) -> BlockSliceQuery:
        """Get block range by cycle/cycle range.

        :param item: Cycle number or range (slice), range start/stop can be empty or negative
        :rtype: BlockSliceQuery
        """
        return self._get_item(item)


class VotingPeriodsQuery(PeriodQuery):
    __pos_key__ = 'voting_period_position'
    __val_key__ = 'voting_period'

    def __call__(self, **params):
        """Get current voting period."""
        return self.head.voting_period()

    def __getitem__(self, item) -> BlockSliceQuery:
        """Get block range by voting_period/voting_period range.

        :param item: Voting_period number or range (slice), range start/stop can be empty or negative
        :rtype: BlockSliceQuery
        """
        return self._get_item(item)



import binascii
import hashlib
import json
from getpass import getpass
from os import environ as env
from os.path import abspath
from os.path import expanduser
from os.path import join
from typing import List
from typing import Optional
from typing import Union

from mnemonic import Mnemonic
from pyblake2 import blake2b  # type: ignore

from pytezos.crypto.encoding import base58_decode
from pytezos.crypto.encoding import base58_encode
from pytezos.crypto.encoding import scrub_input
from pytezos.jupyter import InlineDocstring
from pytezos.jupyter import get_class_docstring

VALID_MNEMONIC_LENGTHS = [12, 15, 18, 21, 24]
DEFAULT_LANGUAGE = 'english'
DEFAULT_TEZOS_DIR = '~/.tezos-client'

PassphraseInput = Optional[Union[str, bytes]]


def get_passphrase(passphrase: PassphraseInput = None, alias: Optional[str] = None) -> bytes:
    if passphrase is None:
        passphrase = env.get('PYTEZOS_PASSPHRASE')
    if passphrase is None:
        passphrase = getpass(f'Please enter passphrase{" for key `" + alias + "`" if alias else ""}:\n')
    if passphrase is None:
        raise Exception('Key passphrase is required')
    if isinstance(passphrase, str):
        passphrase = passphrase.encode()
    if not isinstance(passphrase, bytes):
        raise Exception(f'bytes or str expected, got {type(passphrase).__name__}')
    return passphrase


class CryptoExtraFallback:
    def __getattr__(self, item):
        raise ImportError(
            "Please, install packages libsodium-dev, libsecp256k1-dev, and libgmp-dev, "
            "and Python libraries pysodium, secp256k1, and fastecdsa"
        )

    def __call__(self, *args, **kwargs):
        self.__getattr__('throw')


try:
    import fastecdsa.curve  # type: ignore
    import fastecdsa.ecdsa  # type: ignore
    import fastecdsa.encoding.sec1  # type: ignore
    import fastecdsa.keys  # type: ignore
    import pysodium  # type: ignore
    import secp256k1  # type: ignore
    from fastecdsa.encoding.util import bytes_to_int  # type: ignore
except ImportError as e:
    pysodium = CryptoExtraFallback()
    secp256k1 = CryptoExtraFallback()
    fastecdsa = CryptoExtraFallback()
    bytes_to_int = CryptoExtraFallback()
    __crypto__ = False
else:
    __crypto__ = True


def is_installed():
    return __crypto__


def blake2b_32(v=b''):
    """Get a BLAKE2B hash of bytes"""
    return blake2b(scrub_input(v), digest_size=32)


def validate_mnemonic(mnemonic: str, language: str = DEFAULT_LANGUAGE) -> None:
    m = Mnemonic(language)
    mnemonic_words = m.normalize_string(mnemonic).split(' ')
    if len(mnemonic_words) not in VALID_MNEMONIC_LENGTHS:
        raise ValueError(
            'Number of words must be one of the following: {VALID_MNEMONIC_LENGTHS}, '
            'but it is not (%d).' % len(mnemonic_words)
        )

    idx = map(lambda x: bin(m.wordlist.index(x))[2:].zfill(11), mnemonic_words)
    b = ''.join(idx)
    l = len(b)
    d = b[: l // 33 * 32]
    h = b[-l // 33 :]
    nd = binascii.unhexlify(hex(int(d, 2))[2:].rstrip('L').zfill(l // 33 * 8))
    nh = bin(int(hashlib.sha256(nd).hexdigest(), 16))[2:].zfill(256)[: l // 33]
    if h != nh:
        raise ValueError('Mnemonic checksum verification failed')


class Key(metaclass=InlineDocstring):
    """Represents a public or secret key for Tezos. Ed25519, Secp256k1 and P256
    are supported.
    """

    def __init__(
        self,
        public_point: bytes,
        secret_exponent: Optional[bytes] = None,
        curve: bytes = b'ed',
        activation_code: Optional[str] = None,
    ) -> None:
        self.public_point = public_point
        self.secret_exponent = secret_exponent
        self.curve = curve
        self.activation_code = activation_code

    def __repr__(self) -> str:
        res = [
            super().__repr__(),
            '\nPublic key hash',
            self.public_key_hash(),
            '\nHelpers',
            get_class_docstring(self.__class__),
        ]
        return '\n'.join(res)

    @property
    def is_secret(self) -> bool:
        return self.secret_exponent is not None

    @classmethod
    def from_secret_exponent(
        cls,
        secret_exponent: bytes,
        curve=b'ed',
        activation_code=None,
    ) -> 'Key':
        """Creates a key object from a secret exponent.

        :param secret_exponent: secret exponent or seed
        :param curve: b'sp' for Secp251k1, b'p2' for P256/Secp256r1, b'ed' for Ed25519 (default)
        :param activation_code: secret for initializing account balance
        """
        # Ed25519
        if curve == b'ed':
            # Dealing with secret exponent or seed?
            if len(secret_exponent) == 64:
                public_point = pysodium.crypto_sign_sk_to_pk(sk=secret_exponent)
            else:
                public_point, secret_exponent = pysodium.crypto_sign_seed_keypair(seed=secret_exponent)
        # Secp256k1
        elif curve == b'sp':
            sk = secp256k1.PrivateKey(secret_exponent)
            public_point = sk.pubkey.serialize()
        # P256
        elif curve == b'p2':
            pk = fastecdsa.keys.get_public_key(bytes_to_int(secret_exponent), curve=fastecdsa.curve.P256)
            public_point = fastecdsa.encoding.sec1.SEC1Encoder.encode_public_key(pk)
        else:
            raise AssertionError()

        return cls(public_point, secret_exponent, curve=curve, activation_code=activation_code)

    @classmethod
    def from_public_point(
        cls,
        public_point: bytes,
        curve: bytes = b'ed',
    ) -> 'Key':
        """Creates a key object from a public elliptic point.

        :param public_point: elliptic point in the compressed format (see https://tezos.stackexchange.com/a/623/309)
        :param curve: b'sp' for secp251k1, b'p2' for P256/secp256r1, b'ed' for Ed25519 (default)
        """
        return cls(public_point, curve=curve)

    @classmethod
    def from_encoded_key(
        cls,
        key: Union[str, bytes],
        passphrase: PassphraseInput = None,
    ) -> 'Key':
        """Creates a key object from a base58 encoded key.

        :param key: a public or secret key in base58 encoding
        :param passphrase: the passphrase used if the key provided is an encrypted private key,
            if not set value from from PYTEZOS_PASSPHRASE env variable will be used or promted dynamically
        """
        encoded_key = scrub_input(key)

        curve = encoded_key[:2]  # "sp", "p2" "ed"
        if curve not in [b'sp', b'p2', b'ed']:
            raise ValueError("Invalid prefix for a key encoding.")
        if not len(encoded_key) in [54, 55, 88, 98]:
            raise ValueError("Invalid length for a key encoding.")

        encrypted = encoded_key[2:3] == b'e'
        public_or_secret = encoded_key[3:5] if encrypted else encoded_key[2:4]
        if public_or_secret not in [b'pk', b'sk']:
            raise Exception("Invalid prefix for a key encoding.")

        encoded_key = base58_decode(encoded_key)
        is_secret = public_or_secret == b'sk'
        if not is_secret:
            return cls.from_public_point(encoded_key, curve)

        if encrypted:
            passphrase = get_passphrase(passphrase)

            salt, encrypted_sk = encoded_key[:8], encoded_key[8:]
            encryption_key = hashlib.pbkdf2_hmac(
                hash_name="sha512",
                password=passphrase,
                salt=salt,
                iterations=32768,
                dklen=32,
            )
            encoded_key = pysodium.crypto_secretbox_open(
                c=encrypted_sk,
                nonce=b'\000' * 24,
                k=encryption_key,
            )
            del passphrase

        return cls.from_secret_exponent(encoded_key, curve)

    @classmethod
    def generate(
        cls,
        passphrase: str = '',
        curve: bytes = b'ed',
        strength: int = 128,
        language: str = DEFAULT_LANGUAGE,
        export: bool = True,
    ):
        """Generates new key.

        :param passphrase: optional password
        :param curve: b'sp' for secp251k1, b'p2' for P256/secp256r1, b'ed' for Ed25519 (default)
        :param strength: mnemonic strength, default is 128
        :param language: mnemonic language, default is english
        :param export: export as json file in the current folder, default is True
        :rtype: Key
        """
        mnemonic = Mnemonic(language).generate(strength)
        key = cls.from_mnemonic(mnemonic, passphrase, curve=curve)

        if export:
            pkh = key.public_key_hash()
            data = {
                'mnemonic': mnemonic.split(),
                'pkh': pkh,
                'password': passphrase or '',
            }
            with open(abspath(f'./{pkh}.json'), 'w+') as f:
                f.write(json.dumps(data))

        return key

    @classmethod
    def from_mnemonic(
        cls,
        mnemonic: Union[List[str], str],
        passphrase: str = '',
        email: str = '',
        validate: bool = True,
        curve: bytes = b'ed',
        activation_code: Optional[str] = None,
        language: str = DEFAULT_LANGUAGE,
    ) -> 'Key':
        """Creates a key object from a bip39 mnemonic.

        :param mnemonic: a 15 word bip39 english mnemonic
        :param passphrase: a mnemonic password or a fundraiser key
        :param email: email used if a fundraiser key is passed
        :param validate: whether to check mnemonic or not
        :param curve: b'sp' for secp251k1, b'p2' for P256/secp256r1, b'ed' for Ed25519 (default)
        :param activation_code: secret for initializing account balance
        :param language: The English label for the language of the mnemonic. This is needed for validation
        :rtype: Key
        """
        if isinstance(mnemonic, list):
            mnemonic = ' '.join(mnemonic)

        if validate:
            validate_mnemonic(mnemonic, language=language)

        seed = Mnemonic.to_seed(mnemonic, passphrase=email + passphrase)

        if curve == b'ed':
            _, secret_exponent = pysodium.crypto_sign_seed_keypair(seed=seed[:32])
        elif curve == b'sp':
            secret_exponent = seed[:32]
        elif curve == b'p2':
            secret_exponent = seed[:32]
        else:
            raise AssertionError()

        return cls.from_secret_exponent(secret_exponent, curve=curve, activation_code=activation_code)

    @classmethod
    def from_faucet(cls, source: Union[str, dict]) -> 'Key':
        """Import key from a faucet file: https://teztnets.xyz/

        :param source: path to the json file
        :rtype: Key
        """
        if isinstance(source, str):
            with open(expanduser(source), 'r') as f:
                data = json.loads(f.read())
        elif isinstance(source, dict):
            data = source
        else:
            raise ValueError(f'Unexpected source {source}')

        key = cls.from_mnemonic(
            mnemonic=data['mnemonic'],
            passphrase=data.get('password', ''),
            email=data.get('email', ''),
            activation_code=data['activation_code'],
        )
        if key.public_key_hash() != data['pkh']:
            raise ValueError('Failed to import')

        return key

    @classmethod
    def from_alias(
        cls,
        alias: str,
        passphrase: PassphraseInput = None,
        tezos_client_dir: str = DEFAULT_TEZOS_DIR,
    ) -> 'Key':
        """Import secret key from tezos-client keychain.

        :param alias: key alias
        :param passphrase: if key is encrypted (optional)
        :param tezos_client_dir: path to the tezos client directory (default is `~/.tezos-client`)
        :rtype: Key
        """
        path = expanduser(join(tezos_client_dir, 'secret_keys'))
        with open(path, 'r') as f:
            data = json.loads(f.read())

        value = next(x['value'] for x in data if x['name'] == alias)
        prefix, sk = value.split(':', maxsplit=1)

        if prefix == 'encrypted':
            passphrase = get_passphrase(passphrase)
            key = cls.from_encoded_key(sk, passphrase=passphrase)
            del passphrase
        else:
            key = cls.from_encoded_key(sk)

        return key

    def public_key(self) -> str:
        """Creates base58 encoded public key representation.

        :returns: the public key associated with the private key
        """
        return base58_encode(self.public_point, self.curve + b'pk').decode()

    def secret_key(
        self,
        passphrase: PassphraseInput = None,
        ed25519_seed: bool = True,
    ):
        """Creates base58 encoded private key representation.

        :param passphrase: encryption phrase for the private key
        :param ed25519_seed: encode seed rather than full key for ed25519 curve (True by default)
        :returns: the secret key associated with this key, if available
        """
        if not self.secret_exponent:
            raise ValueError("Secret key is undefined")

        if self.curve == b'ed' and ed25519_seed:
            key = pysodium.crypto_sign_sk_to_seed(self.secret_exponent)
        else:
            key = self.secret_exponent

        if passphrase:
            if not ed25519_seed:
                raise NotImplementedError
            if isinstance(passphrase, str):
                passphrase = passphrase.encode()
            assert isinstance(passphrase, bytes), f'expected bytes or str, got {type(passphrase).__name__}'

            salt = pysodium.randombytes(8)
            encryption_key = hashlib.pbkdf2_hmac(
                hash_name="sha512",
                password=passphrase,
                salt=salt,
                iterations=32768,
                dklen=32,
            )
            encrypted_sk = pysodium.crypto_secretbox(msg=key, nonce=b'\000' * 24, k=encryption_key)
            key = salt + encrypted_sk  # we have to combine salt and encrypted key in order to decrypt later
            prefix = self.curve + b'esk'
        else:
            prefix = self.curve + b'sk'

        return base58_encode(key, prefix).decode()

    def public_key_hash(self) -> str:
        """Creates base58 encoded public key hash for this key.

        :returns: the public key hash for this key
        """
        pkh = blake2b(data=self.public_point, digest_size=20).digest()
        prefix = {b'ed': b'tz1', b'sp': b'tz2', b'p2': b'tz3'}[self.curve]
        return base58_encode(pkh, prefix).decode()

    def blinded_public_key_hash(self) -> str:
        """Creates base58 encoded commitment out of activation code (required) and public key hash

        :return: blinded public key hash
        """
        if not self.activation_code:
            raise ValueError("Activation code is undefined")

        pkh = blake2b(data=self.public_point, digest_size=20).digest()
        key = bytes.fromhex(self.activation_code)
        blinded_pkh = blake2b(data=pkh, key=key, digest_size=20).digest()
        return base58_encode(blinded_pkh, b'btz1').decode()

    def sign(self, message: Union[str, bytes], generic: bool = False):
        """Sign a raw sequence of bytes.

        :param message: sequence of bytes, raw format or hexadecimal notation
        :param generic: do not specify elliptic curve if set to True
        :returns: signature in base58 encoding
        """
        encoded_message = scrub_input(message)

        if not self.secret_exponent:
            raise ValueError("Cannot sign without a secret key.")

        # Ed25519
        if self.curve == b"ed":
            digest = pysodium.crypto_generichash(encoded_message)
            signature = pysodium.crypto_sign_detached(digest, self.secret_exponent)
        # Secp256k1
        elif self.curve == b"sp":
            pk = secp256k1.PrivateKey(self.secret_exponent)
            signature = pk.ecdsa_serialize_compact(pk.ecdsa_sign(encoded_message, digest=blake2b_32))
        # P256
        elif self.curve == b"p2":
            r, s = fastecdsa.ecdsa.sign(msg=encoded_message, d=bytes_to_int(self.secret_exponent), hashfunc=blake2b_32)
            signature = r.to_bytes(32, 'big') + s.to_bytes(32, 'big')
        else:
            raise AssertionError()

        if generic:
            prefix = b'sig'
        else:
            prefix = self.curve + b'sig'

        return base58_encode(signature, prefix).decode()

    def verify(self, signature: Union[str, bytes], message: Union[str, bytes]) -> bool:
        """Verify signature, raise exception if it is not valid.

        :param message: sequance of bytes, raw format or hexadecimal notation
        :param signature: a signature in base58 encoding
        :raises: ValueError if signature is not valid
        :returns: True if signature is valid
        """
        encoded_signature = scrub_input(signature)
        encoded_message = scrub_input(message)

        if not self.public_point:
            raise ValueError("Cannot verify without a public key")

        if encoded_signature[:3] != b'sig':  # not generic
            if self.curve != encoded_signature[:2]:  # "sp", "p2" "ed"
                raise ValueError("Signature and public key curves mismatch.")

        decoded_signature = base58_decode(encoded_signature)

        # Ed25519
        if self.curve == b"ed":
            digest = pysodium.crypto_generichash(encoded_message)
            try:
                pysodium.crypto_sign_verify_detached(decoded_signature, digest, self.public_point)
            except ValueError as exc:
                raise ValueError('Signature is invalid.') from exc
        # Secp256k1
        elif self.curve == b"sp":
            pk = secp256k1.PublicKey(self.public_point, raw=True)
            sig = pk.ecdsa_deserialize_compact(decoded_signature)
            if not pk.ecdsa_verify(encoded_message, sig, digest=blake2b_32):
                raise ValueError('Signature is invalid.')
        # P256
        elif self.curve == b"p2":
            pk = fastecdsa.encoding.sec1.SEC1Encoder.decode_public_key(self.public_point, curve=fastecdsa.curve.P256)
            r, s = bytes_to_int(decoded_signature[:32]), bytes_to_int(decoded_signature[32:])
            if not fastecdsa.ecdsa.verify(sig=(r, s), msg=encoded_message, Q=pk, hashfunc=blake2b_32):
                raise ValueError('Signature is invalid.')
        else:
            raise Exception(f'Unknown elliptic curve {self.curve}')  # type: ignore

        return True



from pytezos.michelson.types.base import MichelsonType
from pytezos.michelson.types.big_map import BigMapType
from pytezos.michelson.types.bls import BLS12_381_FrType
from pytezos.michelson.types.bls import BLS12_381_G1Type
from pytezos.michelson.types.bls import BLS12_381_G2Type
from pytezos.michelson.types.chest import ChestKeyType
from pytezos.michelson.types.chest import ChestType
from pytezos.michelson.types.core import BoolType
from pytezos.michelson.types.core import BytesType
from pytezos.michelson.types.core import IntType
from pytezos.michelson.types.core import NatType
from pytezos.michelson.types.core import NeverType
from pytezos.michelson.types.core import StringType
from pytezos.michelson.types.core import UnitType
from pytezos.michelson.types.domain import AddressType
from pytezos.michelson.types.domain import ChainIdType
from pytezos.michelson.types.domain import ContractType
from pytezos.michelson.types.domain import KeyHashType
from pytezos.michelson.types.domain import KeyType
from pytezos.michelson.types.domain import LambdaType
from pytezos.michelson.types.domain import MutezType
from pytezos.michelson.types.domain import SignatureType
from pytezos.michelson.types.domain import TimestampType
from pytezos.michelson.types.domain import TXRAddress
from pytezos.michelson.types.list import ListType
from pytezos.michelson.types.map import MapType
from pytezos.michelson.types.operation import OperationType
from pytezos.michelson.types.option import OptionType
from pytezos.michelson.types.pair import PairType
from pytezos.michelson.types.sapling import SaplingStateType
from pytezos.michelson.types.sapling import SaplingTransactionDeprecatedType
from pytezos.michelson.types.sapling import SaplingTransactionType
from pytezos.michelson.types.set import SetType
from pytezos.michelson.types.sum import OrType
from pytezos.michelson.types.ticket import TicketType



import logging
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type
from typing import cast

from pytezos.context.abstract import AbstractContext
from pytezos.michelson.instructions.base import MichelsonInstruction
from pytezos.michelson.instructions.base import format_stdout
from pytezos.michelson.micheline import MichelineLiteral
from pytezos.michelson.micheline import MichelineSequence
from pytezos.michelson.micheline import MichelsonRuntimeError
from pytezos.michelson.sections import ParameterSection
from pytezos.michelson.sections import StorageSection
from pytezos.michelson.sections import ViewSection
from pytezos.michelson.stack import MichelsonStack
from pytezos.michelson.types import AddressType
from pytezos.michelson.types import ChainIdType
from pytezos.michelson.types import ContractType
from pytezos.michelson.types import KeyHashType
from pytezos.michelson.types import MutezType
from pytezos.michelson.types import NatType
from pytezos.michelson.types import OperationType
from pytezos.michelson.types import OptionType
from pytezos.michelson.types import PairType
from pytezos.michelson.types import TimestampType
from pytezos.michelson.types import UnitType
from pytezos.michelson.types.base import MichelsonType


class AmountInstruction(MichelsonInstruction, prim='AMOUNT'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        amount = context.get_amount()
        res = MutezType.from_value(amount)
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [], [res]))  # type: ignore
        return cls(stack_items_added=1)


class BalanceInstruction(MichelsonInstruction, prim='BALANCE'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        balance = context.get_balance()
        res = MutezType.from_value(balance)
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [], [res]))  # type: ignore
        return cls(stack_items_added=1)


class ChainIdInstruction(MichelsonInstruction, prim='CHAIN_ID'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        chain_id = context.get_chain_id()
        res = ChainIdType.from_value(chain_id)
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [], [res]))  # type: ignore
        return cls(stack_items_added=1)


def get_entrypoint_type(context: AbstractContext, name: str, address=None) -> Optional[Type[MichelsonType]]:
    expr = context.get_parameter_expr(address)
    if expr is None:
        return None
    parameter = ParameterSection.match(expr)
    entrypoints = parameter.list_entrypoints()
    assert name in entrypoints, f'unknown entrypoint {name}'
    return entrypoints[name]


class SelfInstruction(MichelsonInstruction, prim='SELF'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        entrypoint = next(iter(cls.field_names), 'default')
        self_type = get_entrypoint_type(context, entrypoint)
        assert self_type, f'parameter type is not defined'
        self_address = context.get_self_address()
        res_type = ContractType.create_type(args=[self_type])
        res = res_type.from_value(f'{self_address}%{entrypoint}')  # type: ignore
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [], [res]))  # type: ignore
        return cls(stack_items_added=1)


class SelfAddressInstruction(MichelsonInstruction, prim='SELF_ADDRESS'):
    @classmethod
    def execute(cls, stack: 'MichelsonStack', stdout: List[str], context: AbstractContext):
        res = AddressType.from_value(context.get_self_address())
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [], [res]))  # type: ignore
        return cls(stack_items_added=1)


class SenderInstruction(MichelsonInstruction, prim='SENDER'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        sender = context.get_sender()
        res = AddressType.from_value(sender)
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [], [res]))  # type: ignore
        return cls(stack_items_added=1)


class SourceInstruction(MichelsonInstruction, prim='SOURCE'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        source = context.get_source()
        res = AddressType.from_value(source)
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [], [res]))  # type: ignore
        return cls(stack_items_added=1)


class NowInstruction(MichelsonInstruction, prim='NOW'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        now = context.get_now()
        res = TimestampType.from_value(now)
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [], [res]))  # type: ignore
        return cls(stack_items_added=1)


class AddressInstruction(MichelsonInstruction, prim='ADDRESS'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        contract = cast(ContractType, stack.pop1())
        contract.assert_type_in(ContractType)
        res = AddressType.from_value(contract.get_address())
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [contract], [res]))  # type: ignore
        return cls(stack_items_added=1)


class ContractInstruction(MichelsonInstruction, prim='CONTRACT', args_len=1):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        entrypoint = next(iter(cls.field_names), 'default')
        address = cast(AddressType, stack.pop1())
        address.assert_type_in(AddressType)
        entrypoint_type = get_entrypoint_type(context, entrypoint, address=str(address))
        contract_type = ContractType.create_type(args=cls.args)
        try:
            if entrypoint_type is None:
                stdout.append(f'{cls.prim}: skip type checking for {str(address)}')
            else:
                entrypoint_type.assert_type_equal(cls.args[0])
            res = OptionType.from_some(contract_type.from_value(f'{str(address)}%{entrypoint}'))  # type: ignore
        except AssertionError:
            res = OptionType.none(contract_type)
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [address], [res]))  # type: ignore
        return cls(stack_items_added=1)


class ImplicitAccountInstruction(MichelsonInstruction, prim='IMPLICIT_ACCOUNT'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        key_hash = cast(KeyHashType, stack.pop1())
        key_hash.assert_type_equal(KeyHashType)
        res = ContractType.create_type(args=[UnitType]).from_value(str(key_hash))  # type: ignore
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [key_hash], [res]))  # type: ignore
        return cls(stack_items_added=1)


class CreateContractInstruction(MichelsonInstruction, prim='CREATE_CONTRACT', args_len=1):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        sequence = cast(MichelineSequence, cls.args[0])
        assert len(sequence.args) >= 3, f'expected more than 2 sections, got {len(sequence.args)}'
        assert {arg.prim for arg in sequence.args[:3]} == {'parameter', 'storage', 'code'}, f'unexpected sections'
        storage_type = cast(Type[MichelsonType], next(arg.args[0] for arg in sequence.args if arg.prim == 'storage'))

        delegate, amount, initial_storage = cast(Tuple[OptionType, MutezType, MichelsonType], stack.pop3())
        delegate.assert_type_equal(OptionType.create_type(args=[KeyHashType]))
        amount.assert_type_equal(MutezType)
        initial_storage.assert_type_equal(storage_type)

        originated_address = AddressType.from_value(context.get_originated_address())
        context.spend_balance(int(amount))
        origination = OperationType.origination(
            source=context.get_self_address(),
            script=cls.args[0],  # type: ignore
            storage=initial_storage,
            balance=int(amount),
            delegate=None if delegate.is_none() else str(delegate.get_some()),
        )

        stack.push(originated_address)
        stack.push(origination)
        stdout.append(format_stdout(cls.prim, [delegate, amount, initial_storage], [origination, originated_address]))  # type: ignore
        return cls(stack_items_added=2)


class SetDelegateInstruction(MichelsonInstruction, prim='SET_DELEGATE'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        delegate = cast(OptionType, stack.pop1())
        delegate.assert_type_equal(OptionType.create_type(args=[KeyHashType]))

        delegation = OperationType.delegation(
            source=context.get_self_address(),
            delegate=None if delegate.is_none() else str(delegate.get_some()),
        )
        stack.push(delegation)
        stdout.append(format_stdout(cls.prim, [delegate], [delegation]))  # type: ignore
        return cls(stack_items_added=1)


class TransferTokensInstruction(MichelsonInstruction, prim='TRANSFER_TOKENS'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        parameter, amount, destination = cast(Tuple[MichelsonType, MutezType, ContractType], stack.pop3())
        amount.assert_type_equal(MutezType)
        assert isinstance(destination, ContractType), f'expected contract, got {destination.prim}'
        param_type = destination.args[0]
        parameter.assert_type_equal(param_type)

        ep_type = get_entrypoint_type(context, destination.get_entrypoint(), address=destination.get_address())
        if ep_type:
            parameter.assert_type_equal(ep_type, message='destination contract parameter')

        transaction = OperationType.transaction(
            source=context.get_self_address(),
            destination=destination.get_address(),
            amount=int(amount),
            entrypoint=destination.get_entrypoint(),
            value=parameter.to_micheline_value(),
            param_type=param_type,
        )
        stack.push(transaction)
        stdout.append(format_stdout(cls.prim, [parameter, amount, destination], [transaction]))  # type: ignore
        return cls(stack_items_added=1)


class VotingPowerInstruction(MichelsonInstruction, prim='VOTING_POWER'):
    @classmethod
    def execute(cls, stack: 'MichelsonStack', stdout: List[str], context: AbstractContext):
        address = cast(KeyHashType, stack.pop1())
        address.assert_type_equal(KeyHashType)
        res = NatType.from_value(context.get_voting_power(str(address)))
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [address], [res]))  # type: ignore
        return cls(stack_items_added=1)


class TotalVotingPowerInstruction(MichelsonInstruction, prim='TOTAL_VOTING_POWER'):
    @classmethod
    def execute(cls, stack: 'MichelsonStack', stdout: List[str], context: AbstractContext):
        res = NatType.from_value(context.get_total_voting_power())
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [], [res]))  # type: ignore
        return cls(stack_items_added=1)


class LevelInstruction(MichelsonInstruction, prim='LEVEL'):
    @classmethod
    def execute(cls, stack: 'MichelsonStack', stdout: List[str], context: AbstractContext):
        res = NatType.from_value(context.get_level())
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [], [res]))  # type: ignore
        return cls(stack_items_added=1)


class ViewInstruction(MichelsonInstruction, prim='VIEW', args_len=2):
    @classmethod
    def execute(cls, stack: 'MichelsonStack', stdout: List[str], context: AbstractContext):
        input_value, view_address = cast(Tuple[MichelsonType, AddressType], stack.pop2())

        name = cast(Type[MichelineLiteral], cls.args[0]).get_string()
        address: Optional[str] = str(view_address)
        if address == context.get_self_address():
            address = None
        else:
            # FIXME: spawn new context with patched BALANCE and others
            logging.warning('PyTezos does not support external views with BALANCE or other context-dependent opcodes')

        return_ty = cast(Type[MichelsonType], cls.args[1])

        result = context.get_view_result(address=address, name=name)
        if result is not None:
            logging.info('Using patched VIEW result')
            res = OptionType.from_some(return_ty.from_python_object(result))
        else:
            try:
                view_expr = context.get_view_expr(name, address=address)
                if view_expr is None:
                    raise MichelsonRuntimeError(f'Failed to load view {str(view_address)}%{name}')

                view_ty = ViewSection.match(view_expr)
                return_ty.assert_type_equal(view_ty.args[2], message=f'view {name} return type')
            except (MichelsonRuntimeError, AssertionError) as e:
                stdout.append(f'VIEW: {str(e)}')
                res = OptionType.none(return_ty)
            else:
                storage_expr = context.get_storage_value(address)
                storage_ty = StorageSection.match(context.get_storage_expr())
                storage_value = storage_ty.from_micheline_value(storage_expr).item

                parameter = PairType.from_comb([input_value, storage_value])
                view_stack = MichelsonStack([parameter])
                # FIXME: need to patch context.balance
                view_code = cast(MichelineSequence, view_ty.args[3])
                view_code.execute(view_stack, stdout, context)
                if len(view_stack) != 1:
                    raise MichelsonRuntimeError('Expected single item on the stack, got', view_stack)
                res = OptionType.from_some(view_stack.pop1())

        stack.push(res)
        stdout.append(format_stdout(cls.prim, [input_value, view_address], [res]))  # type: ignore
        return cls(stack_items_added=1)


class OpenChestInstruction(MichelsonInstruction, prim='OPEN_CHEST'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        raise NotImplementedError


class MinBlockTimeInstruction(MichelsonInstruction, prim='MIN_BLOCK_TIME'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        res = NatType.from_value(context.get_min_block_time())
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [], [res]))  # type: ignore
        return cls(stack_items_added=1)


class EmitInstruction(MichelsonInstruction, prim='EMIT', args_len=1):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        event_type = cast(Type[MichelsonType], cls.args[0])
        payload = stack.pop1()
        payload.assert_type_equal(event_type)
        tag = cls.field_names[0] if len(cls.field_names) == 1 else ''
        res = OperationType.event(
            source=context.get_self_address(), event_type=event_type, payload=payload.to_micheline_value(), tag=tag
        )
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [payload], [res], arg=f'%{tag}'))  # type: ignore
        return cls(stack_items_added=0)



from contextlib import suppress
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union

import base58
import strict_rfc3339  # type: ignore

from pytezos.crypto.encoding import base58_decode
from pytezos.crypto.encoding import base58_encode
from pytezos.crypto.key import blake2b_32
from pytezos.michelson.tags import prim_tags

prim_int = {v[0]: k for k, v in prim_tags.items()}


def get_tag(args_len: int, annots_len: int) -> bytes:
    tag = min(args_len * 2 + 3 + (1 if annots_len > 0 else 0), 9)
    return bytes([tag])


def read_tag(tag: int) -> Tuple[int, bool]:
    return (tag - 3) // 2, bool((tag - 3) % 2)


def forge_int(value: int) -> bytes:
    """
    Encode signed unbounded integer into the byte form.

    :param value: signed unbounded integer
    """
    res = bytearray()
    i = abs(value)

    res.append((i & 0b00111111) | (0b11000000 if value < 0 else 0b10000000))
    i >>= 6

    while i != 0:
        res.append((i & 0b01111111) | 0b10000000)
        i >>= 7

    res[-1] &= 0b01111111
    return bytes(res)


def forge_int16(value: int) -> bytes:
    return value.to_bytes(2, 'big')


def forge_int32(value: int) -> bytes:
    return value.to_bytes(4, 'big')


def unforge_int(data: bytes) -> (int, int):  # type: ignore
    """Decode signed unbounded integer from bytes.

    :param data: Encoded integer
    :returns: tuple(parsed integer, length in bytes)
    """
    value = 0
    length = 1

    while data[length - 1] & 0b10000000 != 0:
        length += 1

    for i in range(length - 1, 0, -1):
        value <<= 7
        value |= data[i] & 0b01111111

    value <<= 6
    value |= data[0] & 0b00111111

    if (data[0] & 0b01000000) != 0:
        value = -value

    return value, length


def forge_nat(value: int) -> bytes:
    """Encode a number using LEB128 encoding (Zarith).

    :param int value: the value to encode
    :returns: encoded value
    :rtype: bytes
    """
    if value < 0:
        raise ValueError('Value cannot be negative.')

    buf = bytearray()
    more = True

    while more:
        byte = value & 0x7F
        value >>= 7

        if value:
            byte |= 0x80
        else:
            more = False

        buf.append(byte)

    return bytes(buf)


def unforge_chain_id(data: bytes) -> str:
    """Decode chain id from byte form.

    :param data: encoded chain id.
    :returns: base58 encoded chain id
    """
    return base58_encode(data, b'Net').decode()


def unforge_signature(data: bytes) -> str:
    """Decode signature from byte form.

    :param data: encoded signature.
    :returns: base58 encoded signature (generic)
    """
    return base58_encode(data, b'sig').decode()


def forge_bool(value: bool) -> bytes:
    """Encode boolean value into bytes."""
    return b'\xff' if value else b'\x00'


def forge_base58(value: str) -> bytes:
    """Encode base58 string into bytes.

    :param value: base58 encoded value (with checksum)
    """
    return base58_decode(value.encode())


def optimize_timestamp(value: str) -> int:
    """Convert time string to timestamp.

    :param value: RFC3339 time string or timestamp
    """
    assert isinstance(value, str)
    with suppress(strict_rfc3339.InvalidRFC3339Error):
        return int(strict_rfc3339.rfc3339_to_timestamp(value))
    return int(value)


def forge_address(value: str, tz_only=False) -> bytes:
    """Encode address or key hash into bytes.

    :param value: base58 encoded address or key_hash
    :param tz_only: True indicates that it's a key_hash (will be encoded in a more compact form)
    """
    prefix_len = 4 if value.startswith('txr1') else 3
    prefix = value[:prefix_len]
    address = base58.b58decode_check(value)[prefix_len:]

    if prefix == 'tz1':
        res = b'\x00\x00' + address
    elif prefix == 'tz2':
        res = b'\x00\x01' + address
    elif prefix == 'tz3':
        res = b'\x00\x02' + address
    elif prefix == 'KT1':
        res = b'\x01' + address + b'\x00'
    elif prefix == 'txr1':
        res = b'\x02' + address + b'\x00'
    else:
        raise ValueError(f'Can\'t forge address: unknown prefix `{prefix}`')

    return res[1:] if tz_only else res


def unforge_address(data: bytes) -> str:
    """Decode address or key_hash from bytes.

    :param data: encoded address or key_hash
    :returns: base58 encoded address
    """
    tz_prefixes = {
        b'\x00\x00': b'tz1',
        b'\x00\x01': b'tz2',
        b'\x00\x02': b'tz3',
    }

    for bin_prefix, tz_prefix in tz_prefixes.items():
        if data.startswith(bin_prefix):
            return base58_encode(data[2:], tz_prefix).decode()

    if data.startswith(b'\x01') and data.endswith(b'\x00'):
        return base58_encode(data[1:-1], b'KT1').decode()
    elif data.startswith(b'\x02') and data.endswith(b'\x00'):
        return base58_encode(data[1:-1], b'txr1').decode()
    else:
        return base58_encode(data[1:], tz_prefixes[b'\x00' + data[:1]]).decode()


def forge_contract(value: str) -> bytes:
    """Encode a value of contract type (address + optional entrypoint) into bytes.

    :param value: 'tz12345' or 'tz12345%default'
    """
    parts = value.split('%')
    address, entrypoint = (parts[0], parts[1]) if len(parts) == 2 else (parts[0], 'default')
    res = forge_address(address)
    if entrypoint != 'default':
        res += entrypoint.encode()
    return res


def unforge_contract(data: bytes) -> str:
    """Decode contract (address + optional entrypoint) from bytes

    :param data: encoded contract
    :returns: base58 encoded address and entrypoint (if exists) separated by `%`
    """
    res = unforge_address(data[:22])
    if len(data) > 22:
        res += f'%{data[22:].decode()}'
    return res


def forge_public_key(value: str) -> bytes:
    """Encode public key into bytes.

    :param value: public key in in base58 form
    """
    prefix = value[:4]
    res = base58.b58decode_check(value)[4:]

    if prefix == 'edpk':  # noqa: SIM116
        return b'\x00' + res
    elif prefix == 'sppk':
        return b'\x01' + res
    elif prefix == 'p2pk':
        return b'\x02' + res

    raise ValueError(f'Unrecognized key type: #{prefix}')


def unforge_public_key(data: bytes) -> str:
    """Decode public key from byte form.

    :param data: encoded public key.
    :returns: base58 encoded public key
    """
    key_prefix = {
        b'\x00': b'edpk',
        b'\x01': b'sppk',
        b'\x02': b'p2pk',
    }
    return base58_encode(data[1:], key_prefix[data[:1]]).decode()


def forge_array(data: bytes, len_bytes=4) -> bytes:
    """Encode array of bytes (prepend length).

    :param data: list of bytes
    :param len_bytes: number of bytes to store array length
    """
    return len(data).to_bytes(len_bytes, 'big') + data


def unforge_array(data: bytes, len_bytes=4) -> tuple:
    """Decode array of bytes.

    :param data: encoded array
    :param len_bytes: number of bytes to store array length
    :returns: Tuple[list of bytes, array length]
    """
    assert len(data) >= len_bytes, f'not enough bytes to parse array length, wanted {len_bytes}'
    length = int.from_bytes(data[:len_bytes], 'big')
    assert len(data) >= len_bytes + length, f'not enough bytes to parse array body, wanted {length}'
    return data[len_bytes : len_bytes + length], len_bytes + length


def forge_micheline(data: Union[List, Dict]) -> bytes:
    """Encode a Micheline expression into the byte form.

    :param data: Micheline expression
    """
    res = []

    if isinstance(data, list):
        res.append(b'\x02')
        res.append(forge_array(b''.join(map(forge_micheline, data))))

    elif isinstance(data, dict):
        if data.get('prim'):
            args_len = len(data.get('args', []))
            annots_len = len(data.get('annots', []))

            res.append(get_tag(args_len, annots_len))
            res.append(prim_tags[data['prim']])

            if args_len > 0:
                args = b''.join(map(forge_micheline, data['args']))
                if args_len < 3:
                    res.append(args)
                else:
                    res.append(forge_array(args))

            if annots_len > 0:
                res.append(forge_array(' '.join(data['annots']).encode()))
            elif args_len >= 3:
                res.append(b'\x00' * 4)

        elif data.get('bytes') is not None:
            res.append(b'\x0A')
            res.append(forge_array(bytes.fromhex(data['bytes'])))

        elif data.get('int') is not None:
            res.append(b'\x00')
            res.append(forge_int(int(data['int'])))

        elif data.get('string') is not None:
            res.append(b'\x01')
            res.append(forge_array(data['string'].encode()))
        else:
            raise AssertionError(data)
    else:
        raise AssertionError(data)

    return b''.join(res)


def unforge_micheline(data: bytes) -> Union[List, Dict]:
    """Parse Micheline JSON from bytes.

    :param data: Forged Micheline expression
    :returns: Micheline JSON
    """
    ptr = 0

    def unforge_sequence():
        nonlocal ptr
        _, offset = unforge_array(data[ptr:])
        end, res = ptr + offset, []
        ptr += 4
        while ptr < end:
            res.append(unforge())
        assert ptr == end, f'out of sequence boundaries'
        return res

    def unforge_prim_expr(args_len=0, annots=False):
        nonlocal ptr
        prim_tag = data[ptr]
        ptr += 1
        expr = {'prim': prim_int[prim_tag]}

        if 0 < args_len < 3:
            expr['args'] = [unforge() for _ in range(args_len)]
        elif args_len == 3:
            expr['args'] = unforge_sequence()
        else:
            assert args_len == 0, f'unexpected args len {args_len}'

        if annots or args_len == 3:
            value, offset = unforge_array(data[ptr:])
            ptr += offset
            if len(value) > 0:
                expr['annots'] = value.decode().split(' ')

        return expr

    def unforge():
        nonlocal ptr
        tag = data[ptr]
        ptr += 1
        if tag == 0:
            value, offset = unforge_int(data[ptr:])
            ptr += offset
            return {'int': str(value)}
        elif tag == 1:
            value, offset = unforge_array(data[ptr:])
            ptr += offset
            return {'string': value.decode()}
        elif tag == 2:
            return unforge_sequence()
        elif 2 < tag < 10:
            args_len, annots = read_tag(tag)
            return unforge_prim_expr(args_len, annots)
        elif tag == 10:
            value, offset = unforge_array(data[ptr:])
            ptr += offset
            return {'bytes': value.hex()}
        else:
            raise AssertionError(f'unkonwn tag {tag} at position {ptr}')

    result = unforge()
    assert ptr == len(data), f'have not reach EOS (pos {ptr}/{len(data)})'
    return result


def forge_script(script: Dict[str, Any]) -> bytes:
    """Encode an origination script into the byte form.

    :param script: {"code": "$Micheline_expression", "storage": "$Micheline_expression"}
    """
    code = forge_micheline(script['code'])
    storage = forge_micheline(script['storage'])
    return forge_array(code) + forge_array(storage)


def forge_script_expr(packed_key: bytes) -> str:
    data = blake2b_32(packed_key).digest()
    return base58_encode(data, b'expr').decode()



import json
import os
import shutil
import sys
from os.path import dirname
from os.path import join
from tempfile import TemporaryDirectory

import click
from ipykernel.kernelapp import IPKernelApp
from jupyter_client.kernelspec import KernelSpecManager

from michelson_kernel import __version__
from michelson_kernel.kernel import MichelsonKernel


@click.group()
@click.version_option(__version__)
@click.pass_context
def cli(*_args, **_kwargs):
    pass


@cli.command(help='Run Michelson kernel', context_settings={"ignore_unknown_options": True})
@click.argument('args', nargs=-1, type=str)
@click.pass_context
def run(_ctx, args) -> None:
    IPKernelApp.launch_instance(argv=args, kernel_class=MichelsonKernel)


@cli.command(help='Install Michelson kernel in current environment')
@click.pass_context
def install(_ctx) -> None:
    kernel_json = {
        "argv": [sys.executable, "-m", "michelson_kernel", "run", "-f", "{connection_file}"],
        "display_name": "Michelson",
        "language": "michelson",
        "codemirror_mode": "michelson",
    }
    kernel_js_path = join(dirname(__file__), 'kernel.js')

    kernel_spec = KernelSpecManager()

    with TemporaryDirectory() as td:
        # NOTE: Starts off as 700, not user readable
        os.chmod(td, 0o755)
        shutil.copy(kernel_js_path, join(td, 'kernel.js'))
        with open(os.path.join(td, 'kernel.json'), 'w') as f:
            json.dump(kernel_json, f, sort_keys=True)

        kernel_spec.install_kernel_spec(td, 'michelson', prefix=sys.prefix)


@cli.command(help='Remove Michelson kernel from current environment')
@click.pass_context
def remove(_ctx) -> None:
    kernel_spec = KernelSpecManager()
    kernel_spec.remove_kernel_spec('michelson')



from typing import cast

from py_ecc import optimized_bls12_381 as bls12_381
from py_ecc.bls.constants import POW_2_382
from py_ecc.bls.typing import G1Uncompressed
from py_ecc.bls.typing import G2Uncompressed
from py_ecc.fields import optimized_bls12_381_FQ as FQ
from py_ecc.fields import optimized_bls12_381_FQ2 as FQ2

from pytezos.michelson.micheline import parse_micheline_literal
from pytezos.michelson.types.core import BytesType
from pytezos.michelson.types.core import IntType


class BLS12_381_FrType(IntType, prim='bls12_381_fr'):
    modulus = 0x73EDA753299D7D483339D80809A1D80553BDA402FFFE5BFEFFFFFFFF00000001

    def __init__(self, value: int):
        super(BLS12_381_FrType, self).__init__()
        self.value = value

    @staticmethod
    def bytes_to_int(value: bytes) -> int:
        assert len(value) <= 32, f'expected no more than 32 bytes, got {len(value)}'
        return int.from_bytes(value, 'little')

    @classmethod
    def from_value(cls, value: int) -> 'BLS12_381_FrType':
        return cls(value % cls.modulus)

    @classmethod
    def from_python_object(cls, py_obj) -> 'BLS12_381_FrType':
        if isinstance(py_obj, int):
            value = py_obj
        elif isinstance(py_obj, bytes):
            value = cls.bytes_to_int(py_obj)
        elif isinstance(py_obj, str):
            if py_obj.startswith('0x'):
                py_obj = py_obj[2:]
            value = cls.bytes_to_int(bytes.fromhex(py_obj))
        else:
            raise AssertionError(f'unexpected value {py_obj}')
        return cls.from_value(value)

    @classmethod
    def from_micheline_value(cls, val_expr) -> 'IntType':
        value = parse_micheline_literal(
            val_expr,
            {
                'int': int,
                'bytes': lambda x: cls.bytes_to_int(bytes.fromhex(x)),
            },
        )
        return cls.from_value(value)

    def to_micheline_value(self, mode='readable', lazy_diff=False):
        if mode == 'readable':
            return {'int': str(self.value)}
        else:
            return {'bytes': self.value.to_bytes(32, 'little').hex()}

    def to_python_object(self, try_unpack=False, lazy_diff=False, comparable=False):
        assert not comparable, f'{self.prim} is not comparable'
        return super(BLS12_381_FrType, self).to_python_object()


class BLS12_381_G1Type(BytesType, prim='bls12_381_g1'):
    @classmethod
    def from_value(cls, value: bytes):
        assert len(value) == 96, f'expected 98 bytes, got {len(value)}'
        return cls(value)

    @classmethod
    def from_point(cls, point: G1Uncompressed) -> 'BLS12_381_G1Type':
        if bls12_381.is_inf(point):
            x, y = POW_2_382, 0
        else:
            x_pt, y_pt = bls12_381.normalize(point)
            x, y = x_pt.n, y_pt.n
        value = x.to_bytes(48, 'big') + y.to_bytes(48, 'big')
        return cls.from_value(value)

    def to_point(self) -> G1Uncompressed:
        x = int.from_bytes(self.value[:48], 'big')
        y = int.from_bytes(self.value[48:], 'big')
        point = FQ(x), FQ(y), FQ(1)
        return cast(G1Uncompressed, point)

    def to_python_object(self, try_unpack=False, lazy_diff=False, comparable=False):
        assert not comparable, f'{self.prim} is not comparable'
        return super(BLS12_381_G1Type, self).to_python_object()


class BLS12_381_G2Type(BytesType, prim='bls12_381_g2'):
    @classmethod
    def from_value(cls, value: bytes):
        assert len(value) == 192, f'expected 98 bytes, got {len(value)}'
        return cls(value)

    @classmethod
    def from_point(cls, point: G2Uncompressed) -> 'BLS12_381_G2Type':
        if bls12_381.is_inf(point):
            x_re, x_im, y_re, y_im = 0, POW_2_382, 0, 0
        else:
            x, y = bls12_381.normalize(point)
            x_re, x_im = x.coeffs
            y_re, y_im = y.coeffs
        value = (
            x_im.to_bytes(48, 'big') + x_re.to_bytes(48, 'big') + y_im.to_bytes(48, 'big') + y_re.to_bytes(48, 'big')
        )
        return cls(value)

    def to_point(self) -> G2Uncompressed:
        x_im = int.from_bytes(self.value[:48], 'big')
        x_re = int.from_bytes(self.value[48:96], 'big')
        y_im = int.from_bytes(self.value[96:144], 'big')
        y_re = int.from_bytes(self.value[144:192], 'big')
        point = FQ2([x_re, x_im]), FQ2([y_re, y_im]), FQ2([1, 0])
        return cast(G2Uncompressed, point)

    def to_python_object(self, try_unpack=False, lazy_diff=False, comparable=False):
        assert not comparable, f'{self.prim} is not comparable'
        return super(BLS12_381_G2Type, self).to_python_object()



from typing import Any
from typing import Dict
from typing import List

from pytezos.michelson.forge import forge_array
from pytezos.michelson.forge import forge_base58
from pytezos.michelson.forge import optimize_timestamp


def bump_fitness(fitness: List[str]) -> List[str]:
    if len(fitness) == 0:
        version = 2
        level = 1
        tail = ['', 'ffffffff', '00000000']
    else:
        version = int.from_bytes(bytes.fromhex(fitness[0]), 'big')
        level = int.from_bytes(bytes.fromhex(fitness[1]), 'big') + 1
        tail = ['', 'ffffffff', '00000000']  # locked_round, neg_predecessor_round, round
    return [version.to_bytes(1, 'big').hex(), level.to_bytes(4, 'big').hex(), *tail]


def forge_int_fixed(value: int, length: int) -> bytes:
    return value.to_bytes(length, 'big')


def forge_command(command: str) -> bytes:
    if command == 'activate':
        return b'\x00'
    raise NotImplementedError(command)


def forge_fitness(fitness: List[str]) -> bytes:
    return forge_array(b''.join(map(lambda x: forge_array(bytes.fromhex(x)), fitness)))


def forge_content(content: Dict[str, Any]) -> bytes:
    res = b''
    res += forge_command(content['command'])
    res += forge_base58(content['hash'])
    res += forge_fitness(content['fitness'])
    res += bytes.fromhex(content['protocol_parameters'])
    return res


def forge_protocol_data(protocol_data: Dict[str, Any]) -> bytes:
    res = b''
    if protocol_data.get('content'):
        res += forge_content(protocol_data['content'])
    else:
        res += forge_base58(protocol_data['payload_hash'])
        res += forge_int_fixed(protocol_data['payload_round'], 4)
        res += bytes.fromhex(protocol_data['proof_of_work_nonce'])
        if protocol_data.get('seed_nonce_hash'):
            res += b'\xFF'
            res += forge_base58(protocol_data['seed_nonce_hash'])
        else:
            res += b'\x00'
        if protocol_data.get('liquidity_baking_toggle_vote'):
            res += b'\x01' if protocol_data['liquidity_baking_toggle_vote'] else b'\x00'
        else:
            res += b'\xFF' if protocol_data['liquidity_baking_escape_vote'] else b'\x00'

    return res


def forge_block_header(shell_header: Dict[str, Any]) -> bytes:
    res = forge_int_fixed(shell_header['level'], 4)
    res += forge_int_fixed(shell_header['proto'], 1)
    res += forge_base58(shell_header['predecessor'])
    res += forge_int_fixed(optimize_timestamp(shell_header['timestamp']), 8)
    res += forge_int_fixed(shell_header['validation_pass'], 1)
    res += forge_base58(shell_header['operations_hash'])
    res += forge_fitness(shell_header['fitness'])
    res += forge_base58(shell_header['context'])
    res += bytes.fromhex(shell_header['protocol_data'])
    return res



# Inspired by https://github.com/jansorg/tezos-intellij/blob/master/grammar/michelson.bnf
import json
import re
from typing import List
from typing import Optional

from ply.lex import Lexer  # type: ignore
from ply.lex import LexToken
from ply.lex import lex
from ply.yacc import yacc  # type: ignore

from pytezos.michelson.macros import expand_macro
from pytezos.michelson.macros import expr as make_expr
from pytezos.michelson.tags import prim_tags


def doc(docstring):
    def decorate(func):
        func.__doc__ = docstring
        return func

    return decorate


class MichelsonParserError(ValueError):
    def __init__(self, token: LexToken, message=None):
        message = message or f'failed to parse expression {token}'
        super(MichelsonParserError, self).__init__(message)
        self.message = message
        self.line = token.lineno
        self.pos = token.lexpos

    def format_stdout(self) -> str:
        return f'{self.line}:{self.pos}: {self.message}'


class Sequence(list):
    pass


class SimpleMichelsonLexer(Lexer):
    tokens = ('INT', 'BYTE', 'STR', 'ANNOT', 'PRIM', 'LEFT_CURLY', 'RIGHT_CURLY', 'LEFT_PAREN', 'RIGHT_PAREN', 'SEMI')

    t_INT = r'-?[0-9]+'
    t_BYTE = r'0x[A-Fa-f0-9]*'
    t_STR = r'\"(\\.|[^\"])*\"'
    t_ANNOT = r'[:@%]+([_0-9a-zA-Z\.]*)?'  # r'[:@%]+([_a-zA-Z][_0-9a-zA-Z\.]*)?'
    t_PRIM = r'[A-Za-z][A-Za-z0-9_]+'
    t_LEFT_CURLY = r'\{'
    t_RIGHT_CURLY = r'\}'
    t_LEFT_PAREN = r'\('
    t_RIGHT_PAREN = r'\)'
    t_SEMI = r';'

    t_ignore_MULTI_COMMENT = r'/\*[^*]*\*/'
    t_ignore_COMMENT = r'#[^\n]*'
    t_ignore = ' \t\r\n\f'

    def __init__(self):
        super(SimpleMichelsonLexer, self).__init__()
        self.lexer = lex(module=self, reflags=re.MULTILINE)

    def t_error(self, t):
        t.type = t.value[0]
        t.value = t.value[0]
        t.lexer.skip(1)
        return t


class MichelsonParser:
    """Customizable Michelson parser"""

    tokens = SimpleMichelsonLexer.tokens

    @doc(
        '''instr : expr 
                  | empty'''
    )
    def p_instr(self, p):
        p[0] = p[1]

    @doc('instr : INT')
    def p_instr_int(self, p):
        p[0] = {'int': p[1]}

    @doc('instr : BYTE')
    def p_instr_byte(self, p):
        p[0] = {'bytes': p[1][2:]}  # strip 0x prefix

    @doc('instr : STR')
    def p_instr_str(self, p):
        p[0] = {'string': json.loads(p[1])}

    @doc('instr : instr SEMI instr')
    def p_instr_list(self, p):
        p[0] = []
        for i in [p[1], p[3]]:
            if type(i) is list:
                p[0].extend(i)
            elif i is not None:
                p[0].append(i)

    @doc('instr : LEFT_CURLY instr RIGHT_CURLY')
    def p_instr_subseq(self, p):
        p[0] = Sequence()
        if type(p[2]) is list:
            p[0].extend(p[2])
        elif p[2] is not None:
            p[0].append(p[2])

    @doc('expr : PRIM annots args')
    def p_expr(self, p):
        prim = p[1]
        if prim in prim_tags or prim in self.extra_primitives:
            expr = make_expr(
                prim=prim,
                annots=p[2] or [],
                args=p[3] or [],
            )
        else:
            try:
                expr = expand_macro(
                    prim=prim,
                    annots=p[2] or [],
                    args=p[3] or [],
                )
            except AssertionError as e:
                raise MichelsonParserError(p.slice[1], str(e)) from e
        p[0] = Sequence(expr) if isinstance(expr, list) else expr

    @doc(
        '''annots : annot 
                   | empty'''
    )
    def p_annots(self, p):
        if p[1] is not None:
            p[0] = [p[1]]

    @doc('annots : annots annot')
    def p_annots_list(self, p):
        p[0] = []
        if type(p[1]) == list:
            p[0].extend(p[1])
        if p[2] is not None:
            p[0].append(p[2])

    @doc('annot : ANNOT')
    def p_annot(self, p):
        p[0] = p[1]

    @doc(
        '''args : arg 
                 | empty'''
    )
    def p_args(self, p):
        p[0] = []
        if p[1] is not None:
            p[0].append(p[1])

    @doc('args : args arg')
    def p_args_list(self, p):
        p[0] = []
        if type(p[1]) == list:
            p[0].extend(p[1])
        if p[2] is not None:
            p[0].append(p[2])

    @doc('arg : PRIM')
    def p_arg_prim(self, p):
        p[0] = {'prim': p[1]}

    @doc('arg : INT')
    def p_arg_int(self, p):
        p[0] = {'int': p[1]}

    @doc('arg : BYTE')
    def p_arg_byte(self, p):
        p[0] = {'bytes': p[1][2:]}  # strip 0x prefix

    @doc('arg : STR')
    def p_arg_str(self, p):
        p[0] = {'string': json.loads(p[1])}

    @doc('arg : LEFT_CURLY instr RIGHT_CURLY')
    def p_arg_subseq(self, p):
        if type(p[2]) == list:
            p[0] = p[2]
        elif p[2] is not None:
            p[0] = [p[2]]
        else:
            p[0] = []

    @doc('arg : LEFT_PAREN expr RIGHT_PAREN')
    def p_arg_group(self, p):
        p[0] = p[2]

    @doc('empty :')
    def p_empty(self, p):
        ...

    def p_error(self, p):
        raise MichelsonParserError(p)

    def __init__(self, debug=False, write_tables=False, extra_primitives: Optional[List[str]] = None):
        """Initialize Michelson parser

        :param debug: Verbose output
        :param write_tables: Store PLY output
        :param extra_primitives: List of words to be ignored
        """
        self.lexer = SimpleMichelsonLexer()
        self.parser = yacc(
            module=self,
            debug=debug,
            write_tables=write_tables,
        )
        self.extra_primitives = extra_primitives or []

    def parse(self, code):
        """Parse Michelson source.

        :param code: Michelson source
        :returns: Micheline expression
        """
        if len(code) > 0 and code[0] == '(' and code[-1] == ')':
            code = code[1:-1]
        return self.parser.parse(code)


def michelson_to_micheline(data, parser=None):
    """Converts Michelson source text into a Micheline expression.

    :param data: Michelson string
    :param parser: custom Michelson parser (optional)
    :returns: Micheline expression
    """
    if parser is None:
        parser = MichelsonParser()
    return parser.parse(data)



from typing import Generator
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type
from typing import Union
from typing import cast

from pytezos.context.abstract import AbstractContext
from pytezos.michelson.micheline import Micheline
from pytezos.michelson.types.adt import ADTMixin
from pytezos.michelson.types.adt import Nested
from pytezos.michelson.types.adt import wrap_pair
from pytezos.michelson.types.base import MichelsonType


class PairLiteral(Micheline, prim='Pair', args_len=None):
    pass


class PairType(MichelsonType, ADTMixin, prim='pair', args_len=None):
    def __init__(self, items: Tuple[MichelsonType, ...]):
        super(PairType, self).__init__()
        self.items = items

    def __eq__(self, other):  # type: ignore
        if not isinstance(other, PairType):
            return False
        return all(item == other.items[i] for i, item in enumerate(self.items))

    def __lt__(self, other: 'PairType'):  # type: ignore
        for i, item in enumerate(self.items):  # noqa: SIM111
            if item > other.items[i]:
                return False
        return True

    def __hash__(self):
        return hash(self.items)

    def __repr__(self):
        return f'({" * ".join(map(repr, self.items))})'

    def __iter__(self) -> Generator[MichelsonType, None, None]:
        yield from iter(self.items)

    @classmethod
    def init(cls, items: List[MichelsonType]) -> 'PairType':
        if len(items) > 2:
            right_cls = cast(Type['PairType'], cls.args[1])
            items = items[0], right_cls.init(items[1:])  # type: ignore
        else:
            items = tuple(items)  # type: ignore
        return cls(items)  # type: ignore

    @staticmethod
    def from_comb(items: List[MichelsonType]) -> 'PairType':
        cls = PairType.create_type(args=[type(item) for item in items])
        return cls.init(items)

    @classmethod
    def create_type(
        cls,
        args: List[Type['Micheline']],
        annots: Optional[list] = None,
        **kwargs,
    ) -> Type['PairType']:
        if len(args) > 2:  # comb
            args = [args[0], PairType.create_type(args=args[1:])]
        else:
            assert len(args) == 2, f'unexpected number of args: {len(args)}'
        type_class = super(PairType, cls).create_type(args=args, annots=annots)
        return cast(Type['PairType'], type_class)

    @classmethod
    def iter_type_args(cls, entrypoints=False, path='') -> Generator[Tuple[str, Type[MichelsonType]], None, None]:
        for i, arg in enumerate(cls.args):
            if issubclass(arg, PairType) and not (arg.field_name or arg.type_name):
                yield from arg.iter_type_args(path=path + str(i))  # type: ignore
            else:
                yield path + str(i), arg

    @classmethod
    def generate_pydoc(cls, definitions: list, inferred_name=None, comparable=False):
        name = cls.field_name or cls.type_name or inferred_name or f'{cls.prim}_{len(definitions)}'
        flat_args = cls.get_flat_args(force_tuple=comparable)
        if isinstance(flat_args, dict):
            fields = [(name, arg.generate_pydoc(definitions, inferred_name=name)) for name, arg in flat_args.items()]
            doc = '{\n' + ',\n'.join(f'\t  "{name}": {arg_doc}' for name, arg_doc in fields) + '\n\t}'
        else:
            items = [
                arg.generate_pydoc(definitions, inferred_name=f'{arg.prim}_{i}') for i, arg in enumerate(flat_args)
            ]
            if all(arg.prim in ['pair', 'or'] or not arg.args for arg in flat_args):
                return f'( {", ".join(items)} )'
            else:
                doc = '(\n' + ',\n'.join(f'\t  {arg_doc}' for arg_doc in items) + '\n\t)'
        definitions.insert(0, (name, doc))
        return f'${name}'

    @classmethod
    def dummy(cls, context: AbstractContext) -> 'PairType':
        return cls(tuple(arg.dummy(context) for arg in cls.args))

    @classmethod
    def from_micheline_value(cls, val_expr) -> 'PairType':
        if isinstance(val_expr, dict):
            prim, args = val_expr.get('prim'), val_expr.get('args', [])
            assert prim == 'Pair', f'expected Pair, got {prim}'
        elif isinstance(val_expr, list):
            args = val_expr
        else:
            raise AssertionError(f'either dict(prim) or list expected, got {type(val_expr).__name__}')

        if len(args) == 2:
            value = tuple(cls.args[i].from_micheline_value(arg) for i, arg in enumerate(args))
        elif len(args) > 2:
            value = cls.args[0].from_micheline_value(args[0]), cls.args[1].from_micheline_value(args[1:])
        else:
            raise AssertionError(f'at least two args expected, got {len(args)}')
        return cls(value)

    @classmethod
    def from_python_object(cls, py_obj) -> 'PairType':
        if isinstance(py_obj, list):
            py_obj = tuple(py_obj)

        if isinstance(py_obj, (tuple, dict)):
            path_to_key, key_to_path, idx_to_path = cls.get_type_layout()
            if isinstance(py_obj, tuple):
                py_obj = {idx_to_path[i]: value for i, value in enumerate(py_obj)}
            else:
                assert key_to_path, f'expected named type'
                py_obj = {key_to_path[key]: value for key, value in py_obj.items()}
            try:
                return cls.from_python_object(wrap_pair(py_obj))
            except KeyError as e:
                if not isinstance(path_to_key, dict):
                    path_to_key = {v: f'{k}th' for k, v in idx_to_path.items()}
                field = path_to_key.get(e.args[0], 'some')
                raise KeyError(f'Missing {field} field')
        elif isinstance(py_obj, Nested):
            value = tuple(cls.args[i].from_python_object(py_obj[i]) for i in [0, 1])
            return cls(value)
        else:
            raise AssertionError(f'expected list, tuple, or dict, got {type(py_obj).__name__}')

    def iter_values(self, path='') -> Generator[Tuple[str, MichelsonType], None, None]:
        for i, item in enumerate(self.items):
            if isinstance(item, PairType) and not (item.field_name or item.type_name):
                yield from item.iter_values(path + str(i))
            else:
                yield path + str(i), item

    def iter_comb(self, include_nodes=False) -> Generator[MichelsonType, None, None]:
        if include_nodes:
            yield self
        for i, item in enumerate(self):
            if i == 1 and isinstance(item, PairType) and not (item.field_name or item.type_name):
                yield from item.iter_comb(include_nodes=include_nodes)
            else:
                yield item
                
    def unpairn_comb(self, count) -> Generator[MichelsonType, None, None]:
        for i, item in enumerate(self):
            if i == 1 and isinstance(item, PairType) and not (item.field_name or item.type_name) and count > 0:
                yield from item.unpairn_comb(count - 1)
            else:
                yield item

    def access_comb(self, idx: int) -> MichelsonType:
        return next(item for i, item in enumerate(self.iter_comb(include_nodes=True)) if i == idx)

    def update_comb(self, idx: int, element: MichelsonType) -> 'PairType':
        if idx % 2 == 1:
            leaves = [element if 2 * i + 1 == idx else item for i, item in enumerate(self.iter_comb())]
        else:
            leaves = [item for i, item in enumerate(self.iter_comb()) if 2 * i + 1 < idx]
            if isinstance(element, PairType):
                leaves.extend(element.iter_comb())
            else:
                leaves.append(element)
        return type(self).from_comb(leaves)

    def to_literal(self) -> Type[Micheline]:
        return PairLiteral.create_type(args=[item.to_literal() for item in self.items])

    def to_micheline_value(self, mode='readable', lazy_diff=False):
        if mode == 'legacy_optimized':
            items = self.items
        else:
            items = list(self.iter_comb())
        args = [arg.to_micheline_value(mode=mode, lazy_diff=lazy_diff) for arg in items]
        if mode in ['readable', 'legacy_optimized']:
            return {'prim': 'Pair', 'args': args}
        elif mode == 'optimized':
            if len(args) == 2:
                return {'prim': 'Pair', 'args': args}
            elif len(args) == 3:
                return {'prim': 'Pair', 'args': [args[0], {'prim': 'Pair', 'args': args[1:]}]}
            elif len(args) >= 4:
                return args
            else:
                raise AssertionError(f'unexpected number of args {len(args)}')
        else:
            raise AssertionError(f'unsupported mode {mode}')

    def to_python_object(self, try_unpack=False, lazy_diff=False, comparable=False) -> Union[dict, tuple]:
        flat_values = self.get_flat_values(force_tuple=comparable)
        if isinstance(flat_values, dict):
            return {
                name: arg.to_python_object(
                    try_unpack=try_unpack,
                    lazy_diff=lazy_diff,
                )
                for name, arg in flat_values.items()
            }
        return tuple(
            arg.to_python_object(
                try_unpack=try_unpack,
                lazy_diff=lazy_diff,
                comparable=comparable,
            )
            for arg in flat_values
        )

    def merge_lazy_diff(self, lazy_diff: List[dict]) -> 'PairType':
        items = tuple(item.merge_lazy_diff(lazy_diff) for item in self)
        return type(self)(items)

    def aggregate_lazy_diff(self, lazy_diff: List[dict], mode='readable'):
        items = tuple(item.aggregate_lazy_diff(lazy_diff, mode=mode) for item in self)
        return type(self)(items)

    def attach_context(self, context: AbstractContext, big_map_copy=False):
        for item in self:
            item.attach_context(context, big_map_copy=big_map_copy)

    def __getitem__(self, key: Union[int, str]) -> MichelsonType:
        return self.get_value(key)



import io
import os
import tarfile
from binascii import hexlify
from collections import OrderedDict
from tempfile import TemporaryDirectory
from typing import List
from typing import Tuple

import netstruct  # type: ignore
import requests
import simplejson as json
from tqdm import tqdm  # type: ignore

from pytezos.crypto.encoding import base58_encode
from pytezos.crypto.key import blake2b_32
from pytezos.jupyter import InlineDocstring
from pytezos.jupyter import get_class_docstring
from pytezos.protocol.diff import apply_patch
from pytezos.protocol.diff import generate_unidiff_html
from pytezos.protocol.diff import make_patch


def dir_to_files(path) -> List[Tuple[str, str]]:
    files = []

    with open(os.path.join(path, 'TEZOS_PROTOCOL')) as f:
        index = json.load(f)

    for module in index['modules']:
        for ext in ['mli', 'ml']:
            name = f'{module.lower()}.{ext}'

            filename = os.path.join(path, name)
            if not os.path.exists(filename):
                continue

            with open(filename, 'r') as file:
                text = file.read()
                files.append((name, text))

    return files


def tar_to_files(path=None, raw=None) -> List[Tuple[str, str]]:
    assert path or raw

    fileobj = io.BytesIO(raw) if raw else None
    with tarfile.open(name=path, fileobj=fileobj) as tar, TemporaryDirectory() as tmp_dir:
        tar.extractall(tmp_dir)
        files = dir_to_files(tmp_dir)

    return files


def url_to_files(url) -> List[Tuple[str, str]]:
    res = requests.get(url, stream=True)
    raw = b''

    for data in tqdm(res.iter_content()):
        raw += data

    return tar_to_files(raw=raw)


def files_to_proto(files: List[Tuple[str, str]]) -> dict:
    components = OrderedDict()  # type: ignore

    for filename, text in files:
        name, ext = filename.split('.')
        key = {'mli': 'interface', 'ml': 'implementation'}[ext]
        name = name.capitalize()
        data = hexlify(text.encode()).decode()

        if name in components:
            components[name][key] = data
        else:
            components[name] = {'name': name, key: data}

    proto = {
        'expected_env_version': 0,  # TODO: this is V1
        'components': list(components.values()),
    }
    return proto


def files_to_tar(files: List[Tuple[str, str]], output_path=None):
    fileobj = io.BytesIO() if output_path is None else None
    nameparts = os.path.basename(output_path).split('.')
    mode = 'w'
    if len(nameparts) == 3:
        mode = f'w:{nameparts[-1]}'

    with tarfile.open(name=output_path, fileobj=fileobj, mode=mode) as tar:
        for filename, text in files:
            file = io.BytesIO(text.encode())
            ti = tarfile.TarInfo(filename)
            ti.size = len(file.getvalue())
            tar.addfile(ti, file)

    if fileobj:
        return fileobj.getvalue()
    return None


def proto_to_files(proto: dict) -> List[Tuple[str, str]]:
    files = []
    extensions = {'interface': 'mli', 'implementation': 'ml'}

    for component in proto.get('components', []):
        for key, ext in extensions.items():
            if key in component:
                filename = f'{component["name"].lower()}.{ext}'
                text = bytes.fromhex(component[key]).decode()
                files.append((filename, text))

    return files


def proto_to_bytes(proto: dict) -> bytes:
    res = b''

    for component in proto.get('components', []):
        res += netstruct.pack(b'I$', component['name'].encode())

        if component.get('interface'):
            res += b'\xff' + netstruct.pack(b'I$', bytes.fromhex(component['interface']))
        else:
            res += b'\x00'

        # we should also handle patch case
        res += netstruct.pack(b'I$', bytes.fromhex(component.get('implementation', '')))

    res = netstruct.pack(b'hI$', proto['expected_env_version'], res)
    return res


class Protocol(metaclass=InlineDocstring):
    def __init__(self, proto):
        self._proto = proto

    def __repr__(self):
        res = [
            super().__repr__(),
            '\nHelpers',
            get_class_docstring(self.__class__),
        ]
        return '\n'.join(res)

    def __iter__(self):
        return iter(proto_to_files(self._proto))

    @classmethod
    def from_uri(cls, uri):
        """Loads protocol implementation from various sources and converts it to the RPC-like format.

        :param uri: link/path to a tar archive or path to a folder with extracted contents
        :returns: Protocol instance
        """
        if uri.startswith('http'):
            files = url_to_files(uri)
        elif os.path.exists(os.path.expanduser(uri)):
            files = tar_to_files(uri)
        elif os.path.isdir(uri):
            files = dir_to_files(uri)
        else:
            raise ValueError(uri)

        return Protocol(files_to_proto(files))

    def index(self) -> dict:
        """Generates TEZOS_PROTOCOL file.

        :returns: dict with protocol hash and modules
        """
        data = {
            'hash': self.hash(),
            'modules': list(map(lambda x: x['name'], self._proto.get('components', []))),
        }
        return data

    def export_tar(self, output_path=None):
        """Creates a tarball and dumps to a file or returns bytes.

        :param output_path: Path to the tarball [optional]. You can add .bz2 or .gz extension to make it compressed
        :returns: bytes if path is None or nothing
        """
        files = proto_to_files(self._proto)
        files.append(('TEZOS_PROTOCOL', json.dumps(self.index())))
        return files_to_tar(files, output_path)

    def export_html(self, output_path=None):
        """Generates github-like side-by-side diff viewe, powered by diff2html.js

        :param output_path: will write to this file if specified
        :returns: html string if path is not specified
        """
        diffs = [text for filename, text in self if text]
        return generate_unidiff_html(diffs, output_path=output_path)

    def diff(self, proto, context_size=3):
        """Calculates file diff between two protocol versions.

        :param proto: an instance of Protocol
        :param context_size: number of context lines before and after the change
        :returns: patch in proto format
        """
        files = []
        yours = dict(iter(self))
        theirs = proto_to_files(proto())

        for filename, their_text in theirs:
            patch = make_patch(
                a=yours.get(filename, ''),
                b=their_text,
                filename=filename,
                context_size=context_size,
            )
            files.append((filename, patch))

        return Protocol(files_to_proto(files))

    def patch(self, patch):
        """Applies unified diff and returns full-fledged protocol.

        :param patch: an instance of Protocol containing diff of files
        :returns: Protocol instance
        """
        files = []
        yours = dict(iter(self))
        diff = proto_to_files(patch())

        for filename, diff_text in diff:
            text = yours.get(filename, '')
            if diff_text:
                text = apply_patch(text, diff_text)
            files.append((filename, text))

        return Protocol(files_to_proto(files))

    def hash(self):
        hash_digest = blake2b_32(proto_to_bytes(self._proto)).digest()
        return base58_encode(hash_digest, b'P').decode()



import atexit
import logging
import unittest
from concurrent.futures import FIRST_EXCEPTION
from concurrent.futures import Future
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import wait
from contextlib import suppress
from pprint import pprint
from threading import Event
from time import sleep
from typing import List
from typing import Optional

import requests.exceptions
from testcontainers.core.container import Container  # type: ignore
from testcontainers.core.docker_client import DockerClient  # type: ignore
from testcontainers.core.generic import DockerContainer  # type: ignore

from pytezos.client import PyTezosClient
from pytezos.operation.group import OperationGroup
from pytezos.sandbox.parameters import LATEST
from pytezos.sandbox.parameters import sandbox_addresses

DOCKER_IMAGE = 'bakingbad/sandboxed-node:v14.0-1'
MAX_ATTEMPTS = 100
ATTEMPT_DELAY = 0.1
TEZOS_NODE_PORT = 8732


def kill_existing_containers():
    docker = DockerClient()
    running_containers: List[Container] = docker.client.containers.list(
        filters={
            'status': 'running',
            'ancestor': DOCKER_IMAGE,
        }
    )
    for container in running_containers:
        with suppress(Exception):
            container.stop(timeout=1)


atexit.register(kill_existing_containers)


def worker_callback(f):
    e = f.exception()

    if e is None:
        return

    trace = []
    tb = e.__traceback__
    while tb is not None:
        trace.append(
            {
                "filename": tb.tb_frame.f_code.co_filename,
                "name": tb.tb_frame.f_code.co_name,
                "lineno": tb.tb_lineno,
            }
        )
        tb = tb.tb_next
    pprint(
        {
            'type': type(e).__name__,
            'message': str(e),
            'trace': trace,
        }
    )


def get_next_baker_key(client: PyTezosClient) -> str:
    baking_rights = client.shell.head.helpers.baking_rights()
    delegate = next(br['delegate'] for br in baking_rights if br['round'] == 0)
    return next(k for k, v in sandbox_addresses.items() if v == delegate)


class SandboxedNodeContainer(DockerContainer):
    def __init__(self, image=DOCKER_IMAGE, port=TEZOS_NODE_PORT):
        super().__init__(image)
        self.with_bind_ports(TEZOS_NODE_PORT, port)
        self.url = f'http://localhost:{port}'
        self.client = PyTezosClient().using(shell=self.url)

    def start(self):
        super().start()
        if self.get_wrapped_container() is None:
            raise RuntimeError('Failed to create a container')

    def wait_for_connection(self, max_attempts=MAX_ATTEMPTS, attempt_delay=ATTEMPT_DELAY) -> bool:
        attempts = max_attempts
        while attempts > 0:
            try:
                self.client.shell.node.get("/version/")
                return True
            except requests.exceptions.ConnectionError:
                sleep(attempt_delay)
                attempts -= 1
        return False

    def activate(self, protocol=LATEST):
        return self.client.using(key='dictator').activate_protocol(protocol).fill().sign().inject()

    def bake(self, key: str, min_fee: int = 0):
        return self.client.using(key=key).bake_block(min_fee).fill().work().sign().inject()

    def get_client(self, key: str):
        return self.client.using(key=key)


class SandboxedNodeTestCase(unittest.TestCase):
    """Perform tests with sanboxed node in Docker container."""

    IMAGE: str = DOCKER_IMAGE
    "Docker image to use"

    PORT: int = TEZOS_NODE_PORT
    "Port to expose to host machine"

    PROTOCOL: str = LATEST
    "Hash of protocol to activate"

    node_container: Optional['SandboxedNodeContainer'] = None
    executor: Optional[ThreadPoolExecutor] = None

    @classmethod
    def setUpClass(cls) -> None:
        """Spin up sandboxed node container and activate protocol."""
        kill_existing_containers()
        cls.node_container = SandboxedNodeContainer(image=cls.IMAGE, port=cls.PORT)
        cls.node_container.start()

        if not cls.node_container.wait_for_connection():
            logging.error('failed to connect to %s', cls.node_container.url)
            return

        cls.node_container.activate(cls.PROTOCOL)

    @classmethod
    def tearDownClass(cls) -> None:
        cls._get_node_container().stop(force=True, delete_volume=True)

    @classmethod
    def _get_node_container(cls) -> SandboxedNodeContainer:
        if cls.node_container is None:
            raise RuntimeError('Sandboxed node container is not running')
        return cls.node_container

    @classmethod
    def activate(cls, protocol_alias: str) -> OperationGroup:
        """Activate protocol."""
        return cls._get_node_container().activate(protocol=protocol_alias)

    @classmethod
    def get_client(cls, key='bootstrap2') -> PyTezosClient:
        return cls._get_node_container().get_client(key)

    @classmethod
    def bake_block(cls, min_fee: int = 0) -> OperationGroup:
        """Bake new block.

        :param min_fee: minimum fee of operation to be included in block
        """
        key = get_next_baker_key(cls.get_client())
        return cls._get_node_container().bake(key=key, min_fee=min_fee)

    @property
    def client(self) -> PyTezosClient:
        """PyTezos client to interact with sandboxed node."""
        return self._get_node_container().get_client(key='bootstrap1')


class SandboxedNodeAutoBakeTestCase(SandboxedNodeTestCase):
    exit_event: Optional[Event] = None
    baker: Optional[Future] = None
    min_fee = 0

    TIME_BETWEEN_BLOCKS = 3
    "Time delay between bake attempts, in seconds"

    @staticmethod
    def autobake(time_between_blocks: int, node_url: str, exit_event: Event, min_fee=0):
        logging.info("Baker thread started")
        client = PyTezosClient().using(shell=node_url)
        ptr = 0
        while not exit_event.is_set():
            if ptr % time_between_blocks == 0:
                key = get_next_baker_key(client)
                client.using(key=key).bake_block(min_fee=min_fee).fill().work().sign().inject()
            sleep(1)
            ptr += 1
        logging.info("Baker thread stopped")

    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        if cls.executor is None:
            cls.executor = ThreadPoolExecutor(1)
        if cls.node_container is None:
            raise RuntimeError('sandboxed node container is not created')
        cls.exit_event = Event()
        cls.baker = cls.executor.submit(
            cls.autobake,
            cls.TIME_BETWEEN_BLOCKS,
            cls.node_container.url,
            cls.exit_event,
            cls.min_fee,
        )
        cls.baker.add_done_callback(worker_callback)

    @classmethod
    def tearDownClass(cls) -> None:
        assert cls.exit_event
        assert cls.baker
        cls.exit_event.set()
        wait([cls.baker], return_when=FIRST_EXCEPTION)



import json
import re
from contextlib import suppress
from os.path import dirname
from os.path import join
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import requests
from attr import dataclass
from jsonschema import validate as jsonschema_validate  # type: ignore

from pytezos.context.impl import ExecutionContext
from pytezos.context.mixin import ContextMixin
from pytezos.contract import converter
from pytezos.contract.view import ContractView


def _to_camelcase(string: str) -> str:
    expression = re.compile("[^A-Za-z]+")
    words = expression.split(string)
    if len(words) == 1:
        string = string[0].lower() + string[1:]
        return string
    return ''.join(w.lower() if i == 0 else w.title() for i, w in enumerate(words))


_metadata_json_replace_table = {
    '"return-type":': '"returnType":',
    '"michelson-storage-view":': '"michelsonStorageView":',
}

with open(join(dirname(__file__), 'metadata-schema.json')) as file:
    metadata_schema = json.load(file)


@dataclass(kw_only=True, frozen=True)
class StaticError:
    error: str
    expansion: str
    languages: Optional[List[str]] = None


@dataclass(kw_only=True, frozen=True)
class DynamicError:
    view: str
    languages: Optional[List[str]] = None


@dataclass(kw_only=True, frozen=True)
class License:
    name: str
    details: Optional[str] = None


@dataclass(kw_only=True, frozen=True)
class RestApiView:
    specificationUri: str
    baseUri: Optional[str] = None
    path: str
    method: Optional[str] = None


@dataclass(kw_only=True, frozen=True)
class MichelsonStorageView:
    parameter: Optional[Dict[str, Any]] = None
    returnType: Dict[str, Any]
    code: List[Any]
    annotations: Optional[List[Dict[str, Any]]] = None
    version: Optional[str] = None


@dataclass(kw_only=True, frozen=True)
class RestApiViewImplementation:
    restApiQuery: RestApiView


@dataclass(kw_only=True, frozen=True)
class MichelsonStorageViewImplementation:
    michelsonStorageView: MichelsonStorageView


@dataclass(kw_only=True, frozen=True)
class View:
    name: str
    description: Optional[str] = None
    implementations: List[Union[MichelsonStorageViewImplementation, RestApiViewImplementation]]
    pure: bool = False


@dataclass(kw_only=True, repr=False)
class ContractMetadata(ContextMixin):
    """TZIP-16 contract metadata"""

    name: Optional[str] = None
    description: Optional[str] = None
    version: Optional[str] = None
    license: Optional[License] = None
    authors: Optional[List[str]] = None
    interfaces: Optional[List[str]] = None
    errors: Optional[List[Union[StaticError, DynamicError]]] = None
    views: Optional[List[View]] = None

    def __attrs_post_init__(self) -> None:
        self.raw: Dict[str, Any] = {}
        self.storage_view_impl: Dict[str, Any] = {}
        if self.views:
            for view in self.views:
                michelson_storage_impl = next(
                    (impl for impl in view.implementations if isinstance(impl, MichelsonStorageViewImplementation)),
                    None,
                )
                if michelson_storage_impl is not None:
                    name = _to_camelcase(view.name)
                    self.storage_view_impl[name] = michelson_storage_impl.michelsonStorageView

    def __repr__(self) -> str:
        res = [
            super().__repr__(),
            '\nMetadata',
            f'.name\t\t{self.name or ""}',
            f'.description\t{self.description or ""}',
            f'.license\t{self.license.name if self.license else ""}',
            f'.authors\t{", ".join(self.authors or [])}',
            f'.interfaces\t{", ".join(self.interfaces or [])}',
            f'.errors\t\t<{len(self.errors or [])}>',
            f'.views\t\t<{len(self.views or [])}>',
            '\nStorage views',
            *list(map(lambda x: f'.{x}()', self.storage_view_impl)),
        ]
        return '\n'.join(res)

    def __getattribute__(self, name: str) -> ContractView:
        with suppress(AttributeError):
            return super().__getattribute__(name)
        try:
            impl = self.storage_view_impl[name]
            return ContractView(
                context=self.context,
                name=name,
                parameter=impl.parameter,
                return_type=impl.returnType,
                code=impl.code,
            )
        except KeyError as e:
            raise KeyError(f'Unknown view `{name}`, available: {list(self.storage_view_impl.keys())}') from e

    def __call__(self) -> dict:
        return self.raw

    # FIXME: Unnecessary multiple JSON conversions
    @staticmethod
    def fix_metadata_json(metadata_json: Dict[str, Any]) -> Dict[str, Any]:
        """Fix invalid field casing in metadata JSON"""
        metadata_json_string = json.dumps(metadata_json)
        for from_, to in _metadata_json_replace_table.items():
            metadata_json_string = metadata_json_string.replace(from_, to)
        return json.loads(metadata_json_string)

    @staticmethod
    def validate_metadata_json(metadata_json: Dict[str, Any]) -> None:
        """Validate metadata JSON with JSONSchema"""
        jsonschema_validate(instance=metadata_json, schema=metadata_schema)

    @classmethod
    def from_json(cls, metadata_json: Dict[str, Any], context: Optional[ExecutionContext] = None) -> 'ContractMetadata':
        """Convert metadata from JSON object"""
        metadata_json = cls.fix_metadata_json(metadata_json)
        cls.validate_metadata_json(metadata_json)
        res = converter.structure(metadata_json, ContractMetadata)
        res.context = context if context else ExecutionContext()
        res.raw = metadata_json
        return res

    @classmethod
    def from_ipfs(cls, multihash: str, context: Optional[ExecutionContext] = None) -> 'ContractMetadata':
        """Fetch metadata from IPFS network by multihash"""
        context = context or ExecutionContext()
        metadata_json = requests.get(f'{context.ipfs_gateway}/{multihash}').json()
        return cls.from_json(metadata_json, context)

    @classmethod
    def from_url(cls, url: str, context: Optional[ExecutionContext] = None) -> 'ContractMetadata':
        """Fetch metadata from HTTP(S) URL"""
        metadata_json = requests.get(url).json()
        return cls.from_json(metadata_json, context)

    @classmethod
    def from_file(cls, path: str, context: Optional[ExecutionContext] = None) -> 'ContractMetadata':
        """Read metadata from JSON file by path"""
        with open(path) as f:
            metadata_json = json.load(f)
            return cls.from_json(metadata_json, context)



from typing import List
from typing import Type

from pytezos.context.abstract import AbstractContext
from pytezos.michelson.micheline import Micheline


# FIXME: CodeSection copypaste
class InputSection(Micheline, prim='input', args_len=1):
    @staticmethod
    def match(input_expr) -> Type['InputSection']:
        cls = Micheline.match(input_expr)
        if not issubclass(cls, InputSection):
            cls = InputSection.create_type(args=[cls])
        return cls  # type: ignore

    @classmethod
    def execute(cls, stack, stdout: List[str], context: AbstractContext):
        context.set_input_expr(cls.as_micheline_expr())
        stdout.append('input: updated')


# FIXME: CodeSection copypaste
class OutputSection(Micheline, prim='output', args_len=1):
    @staticmethod
    def match(output_expr) -> Type['OutputSection']:
        cls = Micheline.match(output_expr)
        if not issubclass(cls, OutputSection):
            cls = OutputSection.create_type(args=[cls])
        return cls  # type: ignore

    @classmethod
    def execute(cls, stack, stdout: List[str], context: AbstractContext):
        context.set_output_expr(cls.as_micheline_expr())
        stdout.append('output: updated')


# FIXME: CodeSection copypaste
class SenderSection(Micheline, prim='sender', args_len=1):
    @staticmethod
    def match(output_expr) -> Type['SenderSection']:
        cls = Micheline.match(output_expr)
        if not issubclass(cls, SenderSection):
            cls = SenderSection.create_type(args=[cls])
        return cls  # type: ignore

    @classmethod
    def execute(cls, stack, stdout: List[str], context: AbstractContext):
        context.set_sender_expr(cls.as_micheline_expr())
        stdout.append('sender: updated')


# FIXME: CodeSection copypaste
class BalanceSection(Micheline, prim='balance', args_len=1):
    @staticmethod
    def match(output_expr) -> Type['BalanceSection']:
        cls = Micheline.match(output_expr)
        if not issubclass(cls, BalanceSection):
            cls = BalanceSection.create_type(args=[cls])
        return cls  # type: ignore

    @classmethod
    def execute(cls, stack, stdout: List[str], context: AbstractContext):
        context.set_balance_expr(cls.as_micheline_expr())
        stdout.append('balance: updated')


# FIXME: CodeSection copypaste
class AmountSection(Micheline, prim='amount', args_len=1):
    @staticmethod
    def match(output_expr) -> Type['AmountSection']:
        cls = Micheline.match(output_expr)
        if not issubclass(cls, AmountSection):
            cls = AmountSection.create_type(args=[cls])
        return cls  # type: ignore

    @classmethod
    def execute(cls, stack, stdout: List[str], context: AbstractContext):
        context.set_amount_expr(cls.as_micheline_expr())
        stdout.append('amount: updated')


# FIXME: CodeSection copypaste
class SelfSection(Micheline, prim='self', args_len=1):
    @staticmethod
    def match(output_expr) -> Type['SelfSection']:
        cls = Micheline.match(output_expr)
        if not issubclass(cls, SelfSection):
            cls = SelfSection.create_type(args=[cls])
        return cls  # type: ignore

    @classmethod
    def execute(cls, stack, stdout: List[str], context: AbstractContext):
        context.set_self_expr(cls.as_micheline_expr())
        stdout.append('self: updated')


# FIXME: CodeSection copypaste
class NowSection(Micheline, prim='now', args_len=1):
    @staticmethod
    def match(output_expr) -> Type['NowSection']:
        cls = Micheline.match(output_expr)
        if not issubclass(cls, NowSection):
            cls = NowSection.create_type(args=[cls])
        return cls  # type: ignore

    @classmethod
    def execute(cls, stack, stdout: List[str], context: AbstractContext):
        context.set_now_expr(cls.as_micheline_expr())
        stdout.append('now: updated')


# FIXME: CodeSection copypaste
class SourceSection(Micheline, prim='source', args_len=1):
    @staticmethod
    def match(output_expr) -> Type['SourceSection']:
        cls = Micheline.match(output_expr)
        if not issubclass(cls, SourceSection):
            cls = SourceSection.create_type(args=[cls])
        return cls  # type: ignore

    @classmethod
    def execute(cls, stack, stdout: List[str], context: AbstractContext):
        context.set_source_expr(cls.as_micheline_expr())
        stdout.append('source: updated')


# FIXME: CodeSection copypaste
class ChainIdSection(Micheline, prim='chain_id', args_len=1):
    @staticmethod
    def match(output_expr) -> Type['ChainIdSection']:
        cls = Micheline.match(output_expr)
        if not issubclass(cls, ChainIdSection):
            cls = ChainIdSection.create_type(args=[cls])
        return cls  # type: ignore

    @classmethod
    def execute(cls, stack, stdout: List[str], context: AbstractContext):
        context.set_chain_id_expr(cls.as_micheline_expr())
        stdout.append('chain_id: updated')


# FIXME: CodeSection copypaste
class BigMapsSection(Micheline, prim='big_maps', args_len=1):
    @staticmethod
    def match(output_expr) -> Type['BigMapsSection']:
        cls = Micheline.match(output_expr)
        if not issubclass(cls, BigMapsSection):
            cls = BigMapsSection.create_type(args=[cls])
        return cls  # type: ignore

    @classmethod
    def execute(cls, stack, stdout: List[str], context: AbstractContext):
        context.set_big_maps_expr(cls.as_micheline_expr())
        stdout.append('big_maps: updated')



from typing import Any
from typing import Dict

from pytezos.crypto.key import Key

EDO = 'PtEdo2ZkT9oKpimTah6x2embF25oss54njMuPzkJTEi5RqfdZFA'
FLORENCE = 'PsFLorenaUUuikDWvMDr6fGBRG8kt3e3D3fHoXK1j1BFRxeSH4i'
GRANADA = 'PtGRANADsDU8R9daYKAgWnQYAJ64omN1o3KMGVCykShA97vQbvV'
HANGZHOU = 'PtHangz2aRngywmSRGGvrcTyMbbdpWdpFKuS4uMWxg2RaH9i1qx'
ITHACA = 'Psithaca2MLRFYargivpo7YvUr7wUDqyxrdhC5CQq78mRvimz6A'
JAKARTA = 'PtJakart2xVj7pYXJBXrqHgd82rdkLey5ZeeGwDgPp9rhQUbSqY'
KATHMANDU = 'PtKathmankSpLLDALzWw7CGD2j2MtyveTwboEYokqUCP4a1LxMg'
LATEST = KATHMANDU

protocol_version = {
    EDO: 8,
    FLORENCE: 9,
    GRANADA: 10,
    HANGZHOU: 11,
    ITHACA: 12,
    JAKARTA: 13,
    KATHMANDU: 14,
}

sandbox_commitment = {
    "mnemonic": [
        "arctic",
        "blame",
        "brush",
        "economy",
        "solar",
        "swallow",
        "canvas",
        "live",
        "vote",
        "two",
        "post",
        "neutral",
        "spare",
        "split",
        "fall",
    ],
    "activation_code": "7375ef222cc038001b6c8fb768246c86e994745b",
    "amount": "38323962971",
    "pkh": "tz1W86h1XuWy6awbNUTRUgs6nk8q5vqXQwgk",
    "password": "ZuPOpZgMNM",
    "email": "nbhcylbg.xllfjgrk@tezos.example.org",
}

sandbox_addresses = {
    'activator': 'tz1TGu6TN5GSez2ndXXeDX6LgUDvLzPLqgYV',
    'bootstrap5': 'tz1ddb9NMYHZi5UzPdzTZMYQQZoMub195zgv',
    'bootstrap4': 'tz1b7tUupMgCNw2cCLpKTkSD1NZzB5TkP2sv',
    'bootstrap3': 'tz1faswCTDciRzE4oJ9jn2Vm2dvjeyA9fUzU',
    'bootstrap2': 'tz1gjaF81ZRRvdzjobyfVNsAeSC6PScjfQwN',
    'bootstrap1': 'tz1KqTpEZ7Yob7QbPE4Hy4Wo8fHG8LhKxZSx',
}

sandbox_params: Dict[str, Any] = {
    'bootstrap_accounts': [
        ['edpkuBknW28nW72KG6RoHtYW7p12T6GKc7nAbwYX5m8Wd9sDVC9yav', '4000000000000'],
        ['edpktzNbDAUjUk697W7gYg2CRuBQjyPxbEg8dLccYYwKSKvkPvjtV9', '4000000000000'],
        ['edpkuTXkJDGcFd5nh6VvMz8phXxU3Bi7h6hqgywNFi1vZTfQNnS1RV', '4000000000000'],
        ['edpkuFrRoDSEbJYgxRtLx2ps82UdaYc1WwfS9sE11yhauZt5DgCHbU', '4000000000000'],
        ['edpkv8EUUH68jmo3f7Um5PezmfGrRF24gnfLpH3sVNwJnV5bVCxL2n', '4000000000000'],
    ],
    'bootstrap_contracts': [],
    'commitments': [
        [
            Key.from_faucet(sandbox_commitment).blinded_public_key_hash(),
            '100500000000',
        ],
    ],
    'preserved_cycles': 2.0,
    'blocks_per_cycle': 8.0,
    'blocks_per_commitment': 4.0,
    'cycles_per_voting_period': 64.0,
    'hard_gas_limit_per_operation': '1040000',
    'hard_gas_limit_per_block': '10400000',
    'proof_of_work_threshold': str((1 << 63) - 1),
    'tokens_per_roll': '8000000000',
    'seed_nonce_revelation_tip': '125000',
    'origination_size': 257.0,
    'cost_per_byte': '250',
    'hard_storage_limit_per_operation': '60000',
    'quorum_min': 2000.0,
    'quorum_max': 7000.0,
    'min_proposal_quorum': 500.0,
}


def get_protocol_parameters(protocol_hash: str) -> Dict[str, Any]:
    return {
        **sandbox_params.copy(),
        'cache_sampler_state_cycles': 8.0,
        'tx_rollup_enable': True,
        'tx_rollup_origination_size': 4000.0,
        'liquidity_baking_toggle_ema_threshold': 1000000000.0,
        'cache_script_size': 100000000.0,
        'cache_stake_distribution_cycles': 8.0,
        'tx_rollup_hard_size_limit_per_inbox': 500000.0,
        'tx_rollup_hard_size_limit_per_message': 5000.0,
        'tx_rollup_max_withdrawals_per_batch': 15.0,
        'tx_rollup_commitment_bond': '10000000000',
        'tx_rollup_finality_period': 40000.0,
        'tx_rollup_withdraw_period': 40000.0,
        'tx_rollup_max_inboxes_count': 40100.0,
        'tx_rollup_max_messages_per_inbox': 1010.0,
        'tx_rollup_max_commitments_count': 80100.0,
        'tx_rollup_cost_per_byte_ema_factor': 120.0,
        'tx_rollup_max_ticket_payload_size': 2048.0,
        'tx_rollup_rejection_max_proof_size': 30000.0,
        'tx_rollup_sunset_level': 3473409.0,
        'sc_rollup_enable': False,
        'sc_rollup_origination_size': 6314.0,
        'sc_rollup_challenge_window_in_blocks': 20160.0,
        'sc_rollup_max_available_messages': 1000000.0,
        'sc_rollup_stake_amount': '32000000',
        'sc_rollup_commitment_period_in_blocks': 30.0,
        'sc_rollup_max_lookahead_in_blocks': 30000.0,
        'sc_rollup_max_active_outbox_levels': 20160.0,
        'sc_rollup_max_outbox_messages_per_level': 100.0,
        'blocks_per_stake_snapshot': 4.0,
        'consensus_committee_size': 526.0,
        'consensus_threshold': 0.0,
        'baking_reward_fixed_portion': '333333',
        'baking_reward_bonus_per_slot': '3921',
        'endorsing_reward_per_slot': '2604',
        'max_slashing_period': 2.0,
        'frozen_deposits_percentage': 5.0,
        'max_operations_time_to_live': 120.0,
        'delay_increment_per_round': '1',
        'minimal_participation_ratio': {'numerator': 2.0, 'denominator': 3.0},
        'double_baking_punishment': '640000000',
        'ratio_of_frozen_deposits_slashed_per_double_endorsement': {'numerator': 1.0, 'denominator': 2.0},
        'minimal_block_delay': '1',
        'liquidity_baking_subsidy': '2500000',
        'liquidity_baking_sunset_level': 2032928.0,
        'nonce_revelation_threshold': 2.0,
        'vdf_difficulty': '50000',
        'dal_parametric': {
            'feature_enable': False,
            'number_of_slots': 16.0,
            'number_of_shards': 256.0,
            'endorsement_lag': 1.0,
            'availability_threshold': 50.0,
        },
    }



from pytezos.michelson.sections.code import CodeSection
from pytezos.michelson.sections.parameter import ParameterSection
from pytezos.michelson.sections.storage import StorageSection
from pytezos.michelson.sections.tzt import *
from pytezos.michelson.sections.view import ViewSection



from pytezos.rpc.helpers import *
from pytezos.rpc.node import RpcMultiNode
from pytezos.rpc.node import RpcNode
from pytezos.rpc.protocol import *
from pytezos.rpc.search import *
from pytezos.rpc.shell import *



prim_tags = {
    'parameter': b'\x00',
    'storage': b'\x01',
    'code': b'\x02',
    'False': b'\x03',
    'Elt': b'\x04',
    'Left': b'\x05',
    'None': b'\x06',
    'Pair': b'\x07',
    'Right': b'\x08',
    'Some': b'\x09',
    'True': b'\x0A',
    'Unit': b'\x0B',
    'PACK': b'\x0C',
    'UNPACK': b'\x0D',
    'BLAKE2B': b'\x0E',
    'SHA256': b'\x0F',
    'SHA512': b'\x10',
    'ABS': b'\x11',
    'ADD': b'\x12',
    'AMOUNT': b'\x13',
    'AND': b'\x14',
    'BALANCE': b'\x15',
    'CAR': b'\x16',
    'CDR': b'\x17',
    'CHECK_SIGNATURE': b'\x18',
    'COMPARE': b'\x19',
    'CONCAT': b'\x1A',
    'CONS': b'\x1B',
    '__CREATE_ACCOUNT__': b'\x1C',  # DEPRECATED
    'CREATE_CONTRACT': b'\x1D',
    'IMPLICIT_ACCOUNT': b'\x1E',
    'DIP': b'\x1F',
    'DROP': b'\x20',
    'DUP': b'\x21',
    'EDIV': b'\x22',
    'EMPTY_MAP': b'\x23',
    'EMPTY_SET': b'\x24',
    'EQ': b'\x25',
    'EXEC': b'\x26',
    'FAILWITH': b'\x27',
    'GE': b'\x28',
    'GET': b'\x29',
    'GT': b'\x2A',
    'HASH_KEY': b'\x2B',
    'IF': b'\x2C',
    'IF_CONS': b'\x2D',
    'IF_LEFT': b'\x2E',
    'IF_NONE': b'\x2F',
    'INT': b'\x30',
    'LAMBDA': b'\x31',
    'LE': b'\x32',
    'LEFT': b'\x33',
    'LOOP': b'\x34',
    'LSL': b'\x35',
    'LSR': b'\x36',
    'LT': b'\x37',
    'MAP': b'\x38',
    'MEM': b'\x39',
    'MUL': b'\x3A',
    'NEG': b'\x3B',
    'NEQ': b'\x3C',
    'NIL': b'\x3D',
    'NONE': b'\x3E',
    'NOT': b'\x3F',
    'NOW': b'\x40',
    'OR': b'\x41',
    'PAIR': b'\x42',
    'PUSH': b'\x43',
    'RIGHT': b'\x44',
    'SIZE': b'\x45',
    'SOME': b'\x46',
    'SOURCE': b'\x47',
    'SENDER': b'\x48',
    'SELF': b'\x49',
    'STEPS_TO_QUOTA': b'\x4A',  # DEPRECATED
    'SUB': b'\x4B',
    'SWAP': b'\x4C',
    'TRANSFER_TOKENS': b'\x4D',
    'SET_DELEGATE': b'\x4E',
    'UNIT': b'\x4F',
    'UPDATE': b'\x50',
    'XOR': b'\x51',
    'ITER': b'\x52',
    'LOOP_LEFT': b'\x53',
    'ADDRESS': b'\x54',
    'CONTRACT': b'\x55',
    'ISNAT': b'\x56',
    'CAST': b'\x57',
    'RENAME': b'\x58',
    'bool': b'\x59',
    'contract': b'\x5A',
    'int': b'\x5B',
    'key': b'\x5C',
    'key_hash': b'\x5D',
    'lambda': b'\x5E',
    'list': b'\x5F',
    'map': b'\x60',
    'big_map': b'\x61',
    'nat': b'\x62',
    'option': b'\x63',
    'or': b'\x64',
    'pair': b'\x65',
    'set': b'\x66',
    'signature': b'\x67',
    'string': b'\x68',
    'bytes': b'\x69',
    'mutez': b'\x6A',
    'timestamp': b'\x6B',
    'unit': b'\x6C',
    'operation': b'\x6D',
    'address': b'\x6E',
    'SLICE': b'\x6F',
    'DIG': b'\x70',
    'DUG': b'\x71',
    'EMPTY_BIG_MAP': b'\x72',
    'APPLY': b'\x73',
    'chain_id': b'\x74',
    'CHAIN_ID': b'\x75',
    # EDO
    'LEVEL': b'\x76',
    'SELF_ADDRESS': b'\x77',
    'never': b'\x78',
    'NEVER': b'\x79',
    'UNPAIR': b'\x7A',
    'VOTING_POWER': b'\x7B',
    'TOTAL_VOTING_POWER': b'\x7C',
    'KECCAK': b'\x7D',
    'SHA3': b'\x7E',
    'PAIRING_CHECK': b'\x7F',
    'bls12_381_g1': b'\x80',
    'bls12_381_g2': b'\x81',
    'bls12_381_fr': b'\x82',
    'sapling_state': b'\x83',
    'sapling_transaction_deprecated': b'\x84',
    'SAPLING_EMPTY_STATE': b'\x85',
    'SAPLING_VERIFY_UPDATE': b'\x86',
    'ticket': b'\x87',
    'TICKET': b'\x88',
    'READ_TICKET': b'\x89',
    'SPLIT_TICKET': b'\x8A',
    'JOIN_TICKETS': b'\x8B',
    'GET_AND_UPDATE': b'\x8C',
    # HANGZHOU
    'chest': b'\x8D',
    'chest_key': b'\x8E',
    'OPEN_CHEST': b'\x8F',
    'VIEW': b'\x90',
    'view': b'\x91',
    'constant': b'\x92',
    # ITHACA
    'SUB_MUTEZ': b'\x93',
    # JAKARTA
    'tx_rollup_l2_address': b'\x94',
    'MIN_BLOCK_TIME': b'\x95',
    'sapling_transaction': b'\x96',
    # KATHMANDU
    'EMIT': b'\x97',
    # FIXME: Dummy values for TZT, refactor macros
    'Stack_elt': b'\xEE',
    'Big_map': b'\xEE',
    'input': b'\xEE',
    'output': b'\xEE',
    'sender': b'\xEE',
    'amount': b'\xEE',
    'balance': b'\xEE',
    'self': b'\xEE',
    'now': b'\xEE',
    'source': b'\xEE',
    'big_maps': b'\xEE',
    # FIXME: Dummy values for Jupyter, refactor macros
    'DUMP': b'\xEE',
    'PRINT': b'\xEE',
    'DEBUG': b'\xEE',
    'DROP_ALL': b'\xEE',
    'BEGIN': b'\xEE',
    'COMMIT': b'\xEE',
    'RUN': b'\xEE',
    'EXPAND': b'\xEE',
    'PATCH': b'\xEE',
    'RESET': b'\xEE',
    'BIG_MAP_DIFF': b'\xEE',
}



from pprint import pformat
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from pytezos.context.impl import ExecutionContext
from pytezos.context.mixin import ContextMixin
from pytezos.contract.call import ContractCallResult
from pytezos.contract.call import skip_nones
from pytezos.jupyter import get_class_docstring
from pytezos.logging import logger
from pytezos.michelson.micheline import MichelsonRuntimeError
from pytezos.michelson.repl import Interpreter
from pytezos.michelson.types.base import MichelsonType
from pytezos.michelson.types.base import generate_pydoc


def format_view_script(param_ty_expr, storage_ty_expr, return_ty_expr, code_expr):
    return {
        'code': [
            {
                'prim': 'parameter',
                'args': [
                    {
                        'prim': 'pair',
                        'args': [
                            param_ty_expr,
                            storage_ty_expr['args'][0],
                        ],
                    },
                ],
            },
            {
                'prim': 'storage',
                'args': [
                    {
                        'prim': 'option',
                        'args': [return_ty_expr],
                    },
                ],
            },
            {
                'prim': 'code',
                'args': [
                    [
                        {'prim': 'CAR'},
                        code_expr,
                        {'prim': 'SOME'},
                        {
                            'prim': 'NIL',
                            'args': [
                                {'prim': 'operation'},
                            ],
                        },
                        {'prim': 'PAIR'},
                    ]
                ],
            },
        ]
    }


def format_view_params(param_expr, storage_expr):
    return {
        'entrypoint': 'default',
        'value': {
            'prim': 'Pair',
            'args': [param_expr, storage_expr],
        },
    }


class ContractView(ContextMixin):
    """Proxy class for handling off-chain and on-chain views"""

    def __init__(
        self,
        context: ExecutionContext,
        name: str,
        parameter: Optional[Dict[str, Any]],
        return_type: Dict[str, Any],
        code: List[Any],
    ) -> None:
        super().__init__(context=context)
        self.name = name
        self.param_ty_expr = parameter or {'prim': 'unit'}
        self.return_ty_expr = return_type
        self.code_expr = code
        self.__doc__ = generate_pydoc(MichelsonType.match(self.param_ty_expr), title=name)

    def __repr__(self) -> str:
        res = [
            super().__repr__(),
            f'.name\t{self.name}',
            f'\nBuiltin\n(*args, **kwargs)  # build view parameters (see typedef)',
            f'\nTypedef\n{self.__doc__}',
            '\nHelpers',
            get_class_docstring(self.__class__),
        ]
        return '\n'.join(res)

    def __call__(self, *args, **kwargs) -> 'ContractViewCall':
        """Spawn a contract call proxy initialized with the entrypoint name

        :param args: entrypoint args
        :param kwargs: entrypoint key-value args
        :rtype: ContractViewCall
        """
        if args:
            if len(args) == 1:
                py_obj = args[0]
            else:
                py_obj = args
        elif kwargs:
            py_obj = kwargs
        else:
            py_obj = None

        try:
            param_ty = MichelsonType.match(self.param_ty_expr)
            param_expr = param_ty.from_python_object(py_obj).to_micheline_value()
        except MichelsonRuntimeError as e:
            logger.info(self.__doc__)
            raise ValueError(f'Unexpected arguments: {pformat(py_obj)}', *e.args) from e

        return ContractViewCall(
            context=self.context,
            param_expr=param_expr,
            param_ty_expr=self.param_ty_expr,
            return_ty_expr=self.return_ty_expr,
            code_expr=self.code_expr,
            name=self.name,
        )


class ContractViewCall(ContextMixin):
    """Proxy class encapsulating a contract call: contract type scheme, contract address, parameters, and amount"""

    def __init__(
        self,
        context: ExecutionContext,
        param_expr: dict,
        param_ty_expr: dict,
        return_ty_expr: dict,
        code_expr: list,
        name: str,
    ) -> None:
        super().__init__(context=context)
        self.name = name
        self.param_expr = param_expr
        self.param_ty_expr = param_ty_expr
        self.return_ty_expr = return_ty_expr
        self.code_expr = code_expr

    def _encode_storage(self, storage=None):
        try:
            storage_ty = MichelsonType.match(self.context.storage_expr)
            if storage is None:
                return storage_ty.dummy(self.context).to_micheline_value(lazy_diff=True)
            else:
                return storage_ty.from_python_object(storage).to_micheline_value(lazy_diff=True)
        except MichelsonRuntimeError as e:
            logger.info(self.__doc__)
            raise ValueError(f'Unexpected storage object: {pformat(storage)}', *e.args) from e

    def _get_storage(self, storage=None):
        if storage is None:
            storage_expr = self.context.get_storage_value(self.address)
            if storage_expr is None:
                return self._encode_storage()
            else:
                return storage_expr
        else:
            return self._encode_storage(storage)

    def _prepare_query(
        self,
        parameters=None,
        source=None,
        sender=None,
        balance=None,
        chain_id=None,
        gas_limit=None,
    ):
        script = format_view_script(
            param_ty_expr=self.param_ty_expr,
            storage_ty_expr=self.context.storage_expr,
            return_ty_expr=self.return_ty_expr,
            code_expr=self.code_expr,
        )
        return_ty = MichelsonType.match(self.return_ty_expr)
        initial_storage = return_ty.dummy(self.context).to_micheline_value(lazy_diff=True)
        return skip_nones(
            script=script,
            storage=initial_storage,
            entrypoint=parameters['entrypoint'],
            input=parameters['value'],
            amount='0',
            chain_id=chain_id or self.context.get_chain_id(),
            source=sender,
            payer=source,
            balance=str(balance or 0),
            gas=str(gas_limit) if gas_limit is not None else None,
        )

    def run_code(
        self,
        storage=None,
        source=None,
        sender=None,
        balance=None,
        chain_id=None,
        gas_limit=None,
    ) -> ContractCallResult:
        """Execute using RPC interpreter.

        :param storage: initial storage as Python object, leave None if you want to generate a dummy one
        :param source: patch SOURCE
        :param sender: patch SENDER
        :param balance: patch BALANCE
        :param chain_id: patch CHAIN_ID
        :param gas_limit: restrict max consumed gas
        :rtype: ContractCallResult
        """
        parameters = format_view_params(param_expr=self.param_expr, storage_expr=self._encode_storage(storage))
        query = self._prepare_query(
            source=source, sender=sender, balance=balance, chain_id=chain_id, gas_limit=gas_limit, parameters=parameters
        )
        res = self.shell.blocks[self.block_id].helpers.scripts.run_code.post(query)
        return ContractCallResult.from_run_code(res, parameters=parameters, context=self.context)

    def storage_view(self, storage=None):
        """Get return value of an off-chain view.

        :param storage: override current contract storage (as Python object)
        :returns: Decoded return value
        """
        parameters = format_view_params(
            param_expr=self.param_expr,
            storage_expr=self._get_storage(storage),
        )
        script = format_view_script(
            param_ty_expr=self.param_ty_expr,
            storage_ty_expr=self.context.storage_expr,
            return_ty_expr=self.return_ty_expr,
            code_expr=self.code_expr,
        )
        _, storage, stdout, error = Interpreter.run_callback(
            parameter=parameters['value'],
            entrypoint=parameters['entrypoint'],
            storage={'prim': 'None'},
            context=self._spawn_context(script=script, address=self.address),
        )
        if error:
            logger.debug('\n'.join(stdout))
            raise error
        return storage  # type: ignore

    def onchain_view(self, storage=None, balance=None, view_results: Optional[Dict[str, Any]] = None):
        """Get return value of an on-chain view.

        :param storage: override current contract storage (as Python object)
        :param balance: patch BALANCE
        :param view_results: patch VIEW calls (keys must be string "address%view", values => Python objects)
        :returns: Decoded return value
        """
        ret, stdout, error = Interpreter.run_view(
            name=self.name,
            parameter=self.param_expr,
            storage=self._get_storage(storage),
            context=self._spawn_context(
                balance=balance,
                script={
                    **self.context.script,  # type: ignore
                    'storage': self._encode_storage(storage),
                },
                view_results=view_results,
            ),
        )
        if error:
            logger.debug('\n'.join(stdout))
            raise error
        return ret  # type: ignore



import json
from pprint import pformat
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import requests
import requests.exceptions
from simplejson import JSONDecodeError

from pytezos.logging import logger


def _urljoin(*args: str) -> str:
    return "/".join(map(lambda x: str(x).strip('/'), args))


def _gen_error_variants(error_id: str) -> List[str]:
    chunks = error_id.split('.')
    variants = [error_id]
    if len(chunks) > 1:
        variants.append(chunks[-2])
        if len(chunks) > 2:
            variants.append('.'.join(chunks[2:]))
    return variants


class RpcError(Exception):
    __handlers__ = {}  # type: ignore

    @classmethod
    def __init_subclass__(cls, error_id: Union[str, List[str]]) -> None:
        super().__init_subclass__()
        if isinstance(error_id, list):
            for eid in error_id:
                cls.__handlers__[eid] = cls
        else:
            cls.__handlers__[error_id] = cls

    @classmethod
    def from_errors(cls, errors: List[Dict[str, Any]]) -> 'RpcError':
        """Create RpcError from "errors" section of response JSON."""
        if not errors:
            return RpcError('Unspecified error')

        # FIXME: Only first error is being processed
        error = errors[-1]
        for key in _gen_error_variants(error['id']):
            if key in cls.__handlers__:
                handler = cls.__handlers__[key]
                return handler(error)
        return RpcError(error)

    @classmethod
    def from_response(cls, res: requests.Response) -> 'RpcError':
        """Create RpcError from requests Response."""
        if res.headers.get('content-type') == 'application/json':
            try:
                errors = res.json()
            except JSONDecodeError:
                # sometimes rpc returns invalid json
                return RpcError(res.text)
            assert isinstance(errors, list)
            return cls.from_errors(errors)
        else:
            return RpcError(res.text)

    def __str__(self) -> str:
        return pformat(self.args)


class RpcNode:
    """Request proxy for a single Tezos node."""

    def __init__(self, uri: Union[str, List[str]]) -> None:
        if not uri:
            raise RuntimeError()
        if not isinstance(uri, list):
            uri = [uri]
        self.uri = uri

    def __repr__(self) -> str:
        res = [
            super().__repr__(),
            '\nNode address',
            self.uri[0],
        ]
        return '\n'.join(res)

    def request(self, method: str, path: str, **kwargs) -> requests.Response:
        """Perform HTTP request to node.

        :param method: one of GET/POST/PUT/DELETE
        :param path: path to endpoint
        :param kwargs: requests.request arguments
        :raises RpcError: node has returned an error
        :returns: node response
        """
        logger.debug('>>>>> %s %s\n%s', method, path, json.dumps(kwargs, indent=4))
        res = requests.request(
            method=method,
            url=_urljoin(self.uri[0], path),
            headers={
                'content-type': 'application/json',
                'user-agent': 'PyTezos',
            },
            **kwargs,
        )
        if res.status_code == 401:
            logger.debug('<<<<< %s\n%s', res.status_code, res.text)
            raise RpcError(f'Unauthorized: {path}')
        if res.status_code == 404:
            logger.debug('<<<<< %s\n%s', res.status_code, res.text)
            raise RpcError(f'Not found: {path}')
        if res.status_code != 200:
            logger.debug('<<<<< %s\n%s', res.status_code, pformat(res.text, indent=4))
            raise RpcError.from_response(res)

        logger.debug('<<<<< %s\n%s', res.status_code, json.dumps(res.json(), indent=4))
        return res

    def get(
        self,
        path: str,
        params: Optional[Dict[str, Any]] = None,
        timeout: Optional[int] = None,
    ) -> requests.Response:
        return self.request('GET', path, params=params, timeout=timeout).json()

    def post(self, path: str, params: Optional[Dict[str, Any]] = None, json=None) -> Union[requests.Response, str]:
        response = self.request('POST', path, params=params, json=json)
        try:
            return response.json()
        except JSONDecodeError:
            return response.text

    def delete(self, path: str, params: Optional[Dict[str, Any]] = None) -> requests.Response:
        return self.request('DELETE', path, params=params).json()

    def put(self, path: str, params: Optional[Dict[str, Any]] = None) -> requests.Response:
        return self.request('PUT', path, params=params).json()


class RpcMultiNode(RpcNode):
    """Request proxy for multiple nodes chosen for each request in round-robin order."""

    def __init__(self, uri: Union[str, List[str]]) -> None:
        super().__init__(uri)
        self.nodes = list(map(RpcNode, self.uri))
        self._next_i = 0

    def __repr__(self) -> str:
        res = [
            super().__repr__(),
            '\nNode addresses',
            *self.uri,
        ]
        return '\n'.join(res)

    def request(self, method: str, path: str, **kwargs) -> requests.Response:
        assert self._next_i < len(self.nodes)
        res = self.nodes[self._next_i].request(method, path, **kwargs)
        self._next_i = (self._next_i + 1) % len(self.nodes)
        return res



from typing import Any
from typing import Dict
from typing import Optional

from pytezos.operation.forge import forge_operation

DEFAULT_CONSTANTS = {
    'hard_gas_limit_per_operation': 1040000,
    'hard_storage_limit_per_operation': 60000,
}
# FIXME: Guessed for sandbox tests to pass
DEFAULT_TRANSACTION_GAS_LIMIT = 1451
DEFAULT_TRANSACTION_STORAGE_LIMIT = 257
MINIMAL_FEES = 100
MINIMAL_MUTEZ_PER_BYTE = 1
MINIMAL_MUTEZ_PER_GAS_UNIT = 0.1


def calculate_fee(
    content: Dict[str, Any],
    consumed_gas: int,
    extra_size: int,
    reserve=10,
    minimal_nanotez_per_gas_unit: Optional[int] = None,
) -> int:
    """Calculate minimal required operation fee.

    :param content: operation content {..., "kind": "transaction", ... }
    :param consumed_gas: amount of gas consumed during the simulation (dry-run)
    :param extra_size: size of the additional operation data (branch, etc)
    :param reserve: safe reserve, just in case
    """
    size = len(forge_operation(content)) + extra_size
    if minimal_nanotez_per_gas_unit is None:
        minimal_nanotez_per_gas_unit = int(MINIMAL_MUTEZ_PER_GAS_UNIT * 1000)
    fee = MINIMAL_FEES + MINIMAL_MUTEZ_PER_BYTE * size + int(minimal_nanotez_per_gas_unit * consumed_gas / 1000)
    return fee + reserve


def default_fee(
    content: Dict[str, Any],
    gas_limit: Optional[int] = None,
    minimal_nanotez_per_gas_unit: Optional[int] = None,
) -> int:
    """Take hard gas limit instead of precise amount (no simulation) and calculate fee.

    :param content: operation content {..., "kind": "transaction", ... }
    """
    return calculate_fee(
        content=content,
        consumed_gas=gas_limit if gas_limit is not None else default_gas_limit(content),
        extra_size=32 + 64 + 3 * 3,  # branch, signature, fee:gas_limit:storage_limit mutez values (+3 bytes)
        minimal_nanotez_per_gas_unit=minimal_nanotez_per_gas_unit,
    )


def default_gas_limit(
    content: Dict[str, Any],
    constants: Optional[Dict[str, Any]] = None,
) -> int:
    """Get default gas limit by operation kind.

    :param content: operation content {..., "kind": "transaction", ... }
    :param constants: constants block from context
    """
    if constants is None:
        constants = DEFAULT_CONSTANTS
    values: Dict[str, int] = {
        'reveal': 1000,
        'delegation': 1000,
        'origination': int(constants['hard_gas_limit_per_operation']),
        'transaction': (
            int(constants['hard_gas_limit_per_operation'])
            if content.get('destination', '').startswith('KT')
            else DEFAULT_TRANSACTION_GAS_LIMIT
        ),
        'register_global_constant': int(constants['hard_gas_limit_per_operation']),
    }
    return values[content['kind']]


def default_storage_limit(
    content,
    constants: Optional[Dict[str, Any]] = None,
) -> int:
    """Get default storage limit by operation kind.

    :param content: operation content {..., "kind": "transaction", ... }
    :param constants: constants block from context
    """
    if constants is None:
        constants = DEFAULT_CONSTANTS
    values: Dict[str, int] = {
        'reveal': 0,
        'delegation': 0,
        'origination': int(constants['hard_storage_limit_per_operation']),
        'transaction': (
            int(constants['hard_storage_limit_per_operation'])
            if content.get('destination', '').startswith('KT')
            else DEFAULT_TRANSACTION_STORAGE_LIMIT
        ),
        'register_global_constant': int(constants['hard_storage_limit_per_operation']),
    }
    return values[content['kind']]
